# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from enum import Enum
from types import TracebackType
from typing import Any, Dict, Optional, Tuple, Type, Union

from google.protobuf import text_format
from google.protobuf.message import Message

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest

Exc = Union[
    bool,
    Tuple[Type[BaseException], BaseException, Optional[TracebackType]],
    Tuple[None, None, None],
    BaseException,
]

Tags = Dict[str, Any]


def _str_escape(s: str) -> str:
    return str(s).replace('"', r"\"")


def _format_log_tag_value(value: Any) -> Any:
    if value is None:
        return '""'
    elif isinstance(value, int):
        return value
    elif isinstance(value, float):
        return f"{value:.2f}"
    elif isinstance(value, Digest):
        return f'"{value.hash}/{value.size_bytes}"'
    elif isinstance(value, Message):
        return f'"{_str_escape(text_format.MessageToString(value, as_one_line=True))}"'
    elif isinstance(value, Enum):
        return value.name
    else:
        return f'"{_str_escape(value)}"'


def _format_log_tags(tags: Optional[Tags]) -> str:
    if not tags:
        return ""
    return "".join([f" {key}={_format_log_tag_value(value)}" for key, value in sorted(tags.items())])


class BuildgridLogger:
    def __init__(self, logger: logging.Logger) -> None:
        """
        The buildgrid logger is a helper utility wrapped around a standard logger instance.
        It allows placing key=value strings at the end of log lines, reducing boilerplate in
        displaying values and adding standardization to our log lines. Within each logging method,
        tags may be added by setting the "tags" argument.

        Each logger is set to log at stacklevel=2 such that function names and source line numbers
        show the line at which this utility is invoked.

        Special encoding rules for tag values:
         - int: reported as is. `value=1`
         - float: rounded to the nearest two decimals. `value=1.23`
         - Digest: unpacked as hash/size. `value=deadbeef/123`
         - proto.Message: text_format, escaped, and quoted. `value="blob_digests { hash: \"deadbeef\" }"`
         - Enum: attribute name is used. `value=OK`
         - others: converted to str, escaped, and quoted. `value="foo: \"bar\""`

        Encoding is only performed if logging is enabled for that level.
        """
        self._logger = logger

    def is_enabled_for(self, level: int) -> bool:
        return self._logger.isEnabledFor(level)

    def debug(self, msg: Any, *, exc_info: Optional[Exc] = None, tags: Optional[Tags] = None) -> None:
        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug(str(msg) + _format_log_tags(tags), exc_info=exc_info, stacklevel=2)

    def info(self, msg: Any, *, exc_info: Optional[Exc] = None, tags: Optional[Tags] = None) -> None:
        if self._logger.isEnabledFor(logging.INFO):
            self._logger.info(str(msg) + _format_log_tags(tags), exc_info=exc_info, stacklevel=2)

    def warning(self, msg: Any, *, exc_info: Optional[Exc] = None, tags: Optional[Tags] = None) -> None:
        if self._logger.isEnabledFor(logging.WARNING):
            self._logger.warning(str(msg) + _format_log_tags(tags), exc_info=exc_info, stacklevel=2)

    def error(self, msg: Any, *, exc_info: Optional[Exc] = None, tags: Optional[Tags] = None) -> None:
        if self._logger.isEnabledFor(logging.ERROR):
            self._logger.error(str(msg) + _format_log_tags(tags), exc_info=exc_info, stacklevel=2)

    def exception(self, msg: Any, *, exc_info: Optional[Exc] = True, tags: Optional[Tags] = None) -> None:
        if self._logger.isEnabledFor(logging.ERROR):
            # Note we call error here instead of exception.
            # logger.exception is a helper around calling error with exc_info defaulting to True.
            # On python<3.11 that helper causes the stacklevel to report incorrectly.
            self._logger.error(str(msg) + _format_log_tags(tags), exc_info=exc_info, stacklevel=2)


def buildgrid_logger(name: str) -> BuildgridLogger:
    return BuildgridLogger(logging.getLogger(name))



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import base64
import os
from collections import namedtuple
from typing import Any, Optional, Tuple, TypedDict

import grpc

from buildgrid.server.client.auth_token_loader import AuthTokenLoader
from buildgrid.server.exceptions import InvalidArgumentError

ClientCredentials = TypedDict(
    "ClientCredentials",
    {
        "auth-token": str,
        "tls-client-key": str,
        "tls-client-cert": str,
        "tls-server-cert": str,
        "token-refresh-seconds": int,
    },
    total=False,
)


def load_tls_channel_credentials(
    client_key: Optional[str] = None, client_cert: Optional[str] = None, server_cert: Optional[str] = None
) -> Tuple[grpc.ChannelCredentials, Tuple[Optional[str], Optional[str], Optional[str]]]:
    """Looks-up and loads TLS gRPC client channel credentials.

    Args:
        client_key(str, optional): Client certificate chain file path.
        client_cert(str, optional): Client private key file path.
        server_cert(str, optional): Serve root certificate file path.

    Returns:
        ChannelCredentials: Credentials to be used for a TLS-encrypted gRPC
            client channel.
    """
    if server_cert and os.path.exists(server_cert):
        with open(server_cert, "rb") as f:
            server_cert_pem = f.read()
    else:
        server_cert_pem = None
        server_cert = None

    if client_key and os.path.exists(client_key):
        with open(client_key, "rb") as f:
            client_key_pem = f.read()
    else:
        client_key_pem = None
        client_key = None

    if client_key_pem and client_cert and os.path.exists(client_cert):
        with open(client_cert, "rb") as f:
            client_cert_pem = f.read()
    else:
        client_cert_pem = None
        client_cert = None

    credentials = grpc.ssl_channel_credentials(
        root_certificates=server_cert_pem, private_key=client_key_pem, certificate_chain=client_cert_pem
    )

    return credentials, (
        client_key,
        client_cert,
        server_cert,
    )


class AuthMetadataClientInterceptorBase:
    def __init__(
        self,
        auth_token_loader: Optional[AuthTokenLoader] = None,
        auth_secret: Optional[bytes] = None,
    ) -> None:
        """Initialises a new :class:`AuthMetadataClientInterceptorBase`.

        Important:
            One of `auth_token_path` or `auth_secret` must be provided.

        Args:
            auth_token_loader (AuthTokenLoader, optional): Auth token loader than fetches and passes the token
            auth_secret (bytes, optional): Authorization secret as bytes.

        Raises:
            InvalidArgumentError: If neither `auth_token_loader` or `auth_secret` are
                provided.
        """
        self._auth_token_loader: Optional[AuthTokenLoader] = None
        self.__secret: Optional[str] = None

        if auth_token_loader:
            self._auth_token_loader = auth_token_loader

        elif auth_secret:
            self.__secret = base64.b64encode(auth_secret.strip()).decode()

        else:
            raise InvalidArgumentError("A secret or token must be provided")

        self.__header_field_name = "authorization"

    def _get_secret(self) -> str:
        if self._auth_token_loader:
            token = self._auth_token_loader.get_token()
        else:
            assert self.__secret is not None
            token = self.__secret
        return f"Bearer {token}"

    def amend_call_details(  # type: ignore[no-untyped-def] # wait for client lib updates here
        self, client_call_details, grpc_call_details_class: Any
    ):
        """Appends an authorization field to given client call details."""
        if client_call_details.metadata is not None:
            new_metadata = list(client_call_details.metadata)
        else:
            new_metadata = []

        new_metadata.append(
            (
                self.__header_field_name,
                self._get_secret(),
            )
        )

        class _ClientCallDetails(
            namedtuple(
                "_ClientCallDetails",
                (
                    "method",
                    "timeout",
                    "credentials",
                    "metadata",
                    "wait_for_ready",
                ),
            ),
            grpc_call_details_class,  # type: ignore
        ):
            pass

        return _ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            client_call_details.credentials,
            new_metadata,
            client_call_details.wait_for_ready,
        )



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import itertools
from typing import Any, Callable, Iterator, TypeVar, cast

import grpc

from buildgrid.server.context import instance_context

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]


def instanced(get_instance_name: Callable[[Any], str]) -> Callable[[Func], Func]:
    def decorator(f: Func) -> Func:
        @functools.wraps(f)
        def server_stream_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Iterator[Any]:
            if isinstance(message, Iterator):
                # Pop the message out to get the instance from it, then and recreate the iterator.
                first_message = next(message)
                message = itertools.chain([first_message], message)
                instance_name = get_instance_name(first_message)
            else:
                instance_name = get_instance_name(message)

            with instance_context(instance_name):
                yield from f(self, message, context)

        @functools.wraps(f)
        def server_unary_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Any:
            if isinstance(message, Iterator):
                # Pop the message out to get the instance from it, then and recreate the iterator.
                first_message = next(message)
                message = itertools.chain([first_message], message)
                instance_name = get_instance_name(first_message)
            else:
                instance_name = get_instance_name(message)

            with instance_context(instance_name):
                return f(self, message, context)

        if inspect.isgeneratorfunction(f):
            return cast(Func, server_stream_wrapper)
        return cast(Func, server_unary_wrapper)

    return decorator



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2
from buildgrid._protos.buildgrid.v2.identity_pb2 import ClientIdentity
from buildgrid.server.auth.manager import set_context_client_identity
from buildgrid.server.metadata import extract_client_identity, extract_request_metadata
from buildgrid.server.settings import REQUEST_METADATA_HEADER_NAME


def generate_metadata():
    request_metadata = remote_execution_pb2.RequestMetadata()
    request_metadata.tool_details.tool_name = "tool-name"
    request_metadata.tool_details.tool_version = "1.0"

    request_metadata.action_id = "920ea86d6a445df893d0a39815e8856254392ce40e5957f167af8f16485916fb"
    request_metadata.tool_invocation_id = "cec14b04-075e-4f47-9c24-c5e6ac7f9827"
    request_metadata.correlated_invocations_id = "9f962383-25f6-43b3-886d-56d761ac524e"

    from collections import namedtuple

    metadata_entry = namedtuple("metadata_entry", ("key", "value"))

    return [metadata_entry(REQUEST_METADATA_HEADER_NAME, request_metadata.SerializeToString())]


def test_extract_request_metadata():
    metadata = extract_request_metadata(generate_metadata())

    assert metadata.tool_details.tool_name == "tool-name"
    assert metadata.tool_details.tool_version == "1.0"
    assert metadata.action_id == "920ea86d6a445df893d0a39815e8856254392ce40e5957f167af8f16485916fb"
    assert metadata.tool_invocation_id == "cec14b04-075e-4f47-9c24-c5e6ac7f9827"
    assert metadata.correlated_invocations_id == "9f962383-25f6-43b3-886d-56d761ac524e"


def test_extract_request_metadata_with_no_metadata():
    metadata = extract_request_metadata(())
    assert metadata == remote_execution_pb2.RequestMetadata()


def test_extract_client_identity():
    mock_invocation_metadata = [
        ("x-request-workflow", "workflow"),
        ("x-request-actor", "tool"),
        ("x-request-subject", "user"),
    ]

    client_id = extract_client_identity("instance", mock_invocation_metadata)

    assert client_id.instance == "instance"
    assert client_id.workflow == "workflow"
    assert client_id.actor == "tool"
    assert client_id.subject == "user"


def test_extract_client_identity_empty():
    mock_invocation_metadata = []

    client_id = extract_client_identity("instance", mock_invocation_metadata)

    assert client_id is None


def test_extract_client_identity_with_ctx_var():
    mock_invocation_metadata = [
        ("x-request-workflow", "workflow"),
        ("x-request-actor", "tool"),
        ("x-request-subject", "user"),
    ]

    client_id = extract_client_identity("instance", mock_invocation_metadata)

    assert client_id.instance == "instance"
    assert client_id.workflow == "workflow"
    assert client_id.actor == "tool"
    assert client_id.subject == "user"

    set_context_client_identity(
        ClientIdentity(
            workflow="buildgrid",
            actor="janedoe",
            subject="johndoe",
        )
    )

    client_id = extract_client_identity("instance", mock_invocation_metadata)
    assert client_id.instance == "instance"
    assert client_id.workflow == "buildgrid"
    assert client_id.actor == "janedoe"
    assert client_id.subject == "johndoe"



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Worker(google.protobuf.message.Message):
    """Describes a worker, which is a list of one or more devices and the
    connections between them. A device could be a computer, a phone, or even an
    accelerator like a GPU; it's up to the farm administrator to decide how to
    model their farm. For example, if a farm only has one type of GPU, the GPU
    could be modelled as a "has_gpu" property on its host computer; if it has
    many subproperties itself, it might be better to model it as a separate
    device.

    The first device in the worker is the "primary device" - that is, the device
    running a bot and which is responsible for actually executing commands. All
    other devices are considered to be attached devices, and must be controllable
    by the primary device.

    This message (and all its submessages) can be used in two contexts:

    * Status: sent by the bot to report the current capabilities of the device to
    allow reservation matching.
    * Request: sent by a client to request a device with certain capabilities in
    a reservation.

    Several of the fields in this message have different semantics depending on
    which of which of these contexts it is used. These semantics are described
    below.

    Several messages in Worker and its submessages have the concept of keys and
    values, such as `Worker.Property` and `Device.Property`. All keys are simple
    strings, but certain keys are "standard" keys and should be broadly supported
    across farms and implementations; these are listed below each relevant
    message. Bot implementations or farm admins may add *additional* keys, but
    these SHOULD all begin with an underscore so they do not conflict with
    standard keys that may be added in the future.

    Keys are not context sensitive.

    See http://goo.gl/NurY8g for more information on the Worker message.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class Property(google.protobuf.message.Message):
        """A global property; see the `properties` field for more information."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        """For general information on keys, see the documentation to `Worker`.

        The current set of standard keys are:

        * pool: different workers can be reserved for different purposes. For
        example, an admin might want to segregate long-running integration tests
        from short-running unit tests, so unit tests will always get some
        throughput. To support this, the server can assign different values for
        `pool` (such as "itest" and "utest") to different workers, and then have
        jobs request workers from those pools.
        """
        value: builtins.str
        """The property's value."""
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing_extensions.final
    class Config(google.protobuf.message.Message):
        """A configuration request or report; see the `configs` field for more
        information.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        """For general information on keys, see the documentation to `Worker`.

        The current set of standard keys are:

        * DockerImage: the image of the container. When being reported by the
        bot, the empty value should always be included if the bot is able to pull
        its own images; the bot may optionally *also* report images that are
        present in its cache. When being requested in a lease, the value is the
        URI of the image (eg `gcr.io/user/image@sha256:hash`).
        """
        value: builtins.str
        """The configuration's value."""
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    DEVICES_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    CONFIGS_FIELD_NUMBER: builtins.int
    @property
    def devices(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Device]:
        """A list of devices; the first device is the primary device. See the `Device`
        message for more information.
        """
    @property
    def properties(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Worker.Property]:
        """A worker may contain "global" properties. For example, certain machines
        might be reserved for certain types of jobs, like short-running compilation
        versus long-running integration tests. This property is known as a "pool"
        and is not related to any one device within the worker; rather, it applies
        to the worker as a whole.

        The behaviour of repeated keys is identical to that of Device.Property.
        """
    @property
    def configs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Worker.Config]:
        """Bots can be configured in certain ways when accepting leases. For example,
        many leases are executed inside a Docker container. To support this, the
        bot needs to be able to report that it has Docker installed (and knows how
        to execute something inside a container), and the task submitter needs to
        specify which image should be used to start the container. Similarly, a
        lease may be able to run as one of several users on the worker; in such
        cases, the bot needs to report what users are available, and the submitter
        needs to choose one.

        Therefore, when this message is reported by the bot to the service, each
        key represents a *type* of configuration that the bot knows how to set,
        while each *value* represents a legal value for that configuration (the
        empty string is interpretted as a wildcard, such as for Docker images).
        When this message is sent by the server to the bot in the context of a
        lease, it represents a command to the bot to apply the setting. Keys may
        be repeated during reporting but not in a lease.
        """
    def __init__(
        self,
        *,
        devices: collections.abc.Iterable[global___Device] | None = ...,
        properties: collections.abc.Iterable[global___Worker.Property] | None = ...,
        configs: collections.abc.Iterable[global___Worker.Config] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["configs", b"configs", "devices", b"devices", "properties", b"properties"]) -> None: ...

global___Worker = Worker

@typing_extensions.final
class Device(google.protobuf.message.Message):
    """Any device, including computers, phones, accelerators (e.g. GPUs), etc. All
    names must be unique.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class Property(google.protobuf.message.Message):
        """A device property; see `properties` for more information."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        """For general information on keys, see the documentation to `Worker`.

        The current set of standard keys are:

        * os: a human-readable description of the OS. Examples include `linux`,
        `ubuntu` and `ubuntu 14.04` (note that a bot may advertise itself as more
        than one). This will be replaced in the future by more well-structured
        keys and values to represent OS variants.

        * has-docker: "true" if the bot has Docker installed. This will be
        replaced in the future by a more structured message for Docker support.
        """
        value: builtins.str
        """The property's value."""
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    HANDLE_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    handle: builtins.str
    """The handle can be thought of as the "name" of the device, and must be
    unique within a Worker.

    In the Status context, the handle should be some human-understandable name,
    perhaps corresponding to a label physically written on the device to make
    it easy to locate. In the Request context, the name should be the
    *logical* name expected by the task. The bot is responsible for mapping the
    logical name expected by the task to a machine-readable name that the task
    can actually use, such as a USB address. The method by which this mapping
    is communicated to the task is not covered in this API.
    """
    @property
    def properties(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Device.Property]:
        """Properties of this device that don't change based on the tasks that are
        running on it, e.g. OS, CPU architecture, etc.

        Keys may be repeated, and have the following interpretation:

           * Status context: the device can support *any* the listed values. For
           example, an "ISA" property might include "x86", "x86-64" and "sse4".

           * Request context: the device *must* support *all* of the listed values.
        """
    def __init__(
        self,
        *,
        handle: builtins.str = ...,
        properties: collections.abc.Iterable[global___Device.Property] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["handle", b"handle", "properties", b"properties"]) -> None: ...

global___Device = Device



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
WithCacheStorage
==================

A storage provider that first checks a cache, then tries a slower
fallback provider.

To ensure clients can reliably store blobs in CAS, only `get_blob`
calls are cached -- `has_blob` and `missing_blobs` will always query
the fallback.
"""


from contextlib import ExitStack
from typing import IO, Dict, List, Optional, Tuple

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric, publish_distribution_metric
from buildgrid.server.settings import MAX_IN_MEMORY_BLOB_SIZE_BYTES
from buildgrid.server.threading import ContextThreadPoolExecutor

from .storage_abc import StorageABC, create_write_session

LOGGER = buildgrid_logger(__name__)


class WithCacheStorage(StorageABC):
    TYPE = "WithCache"

    def __init__(
        self,
        cache: StorageABC,
        fallback: StorageABC,
        defer_fallback_writes: bool = False,
        fallback_writer_threads: int = 20,
    ) -> None:
        self._stack = ExitStack()
        self._cache = cache
        self._fallback = fallback
        self._defer_fallback_writes = defer_fallback_writes
        self._fallback_writer_threads = fallback_writer_threads
        self._executor = ContextThreadPoolExecutor(self._fallback_writer_threads, "WithCacheFallbackWriter")

    def start(self) -> None:
        if self._defer_fallback_writes:
            self._stack.enter_context(self._executor)
        self._stack.enter_context(self._cache)
        self._stack.enter_context(self._fallback)

    def stop(self) -> None:
        self._stack.close()
        LOGGER.info(f"Stopped {type(self).__name__}")

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        try:
            if self._defer_fallback_writes and self._cache.has_blob(digest):
                return True
        except Exception:
            LOGGER.warning(
                "Failed to check existence of digest in cache storage.", tags=dict(digest=digest), exc_info=True
            )

        return self._fallback.has_blob(digest)

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        try:
            cache_result = self._cache.get_blob(digest)
            if cache_result is not None:
                publish_distribution_metric(METRIC.STORAGE.WITH_CACHE.CACHE_HIT_COUNT, 1)
                return cache_result
        except Exception:
            LOGGER.warning("Failed to read digest from cache storage.", tags=dict(digest=digest), exc_info=True)

        fallback_result = self._fallback.get_blob(digest)
        if fallback_result is None:
            return None

        publish_distribution_metric(METRIC.STORAGE.WITH_CACHE.CACHE_MISS_COUNT, 1)

        try:
            self._cache.commit_write(digest, fallback_result)
        except Exception:
            LOGGER.warning(
                "Failed to write digest to cache storage after reading blob.", tags=dict(digest=digest), exc_info=True
            )

        fallback_result.seek(0)
        return fallback_result

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        self._fallback.delete_blob(digest)
        try:
            self._cache.delete_blob(digest)
        except Exception:
            LOGGER.warning("Failed to delete digest from cache storage.", tags=dict(digest=digest), exc_info=True)

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        # Only report back failures from the fallback
        try:
            cache_failures = self._cache.bulk_delete(digests)
            for failure in cache_failures:
                LOGGER.warning("Failed to delete digest from cache storage.", tags=dict(digest=failure))
            publish_counter_metric(METRIC.STORAGE.DELETE_ERRORS_COUNT, len(cache_failures), type=self.TYPE)
        except Exception:
            LOGGER.warning("Failed to bulk delete blobs from cache storage.", exc_info=True)

        fallback_failures = self._fallback.bulk_delete(digests)
        publish_counter_metric(METRIC.STORAGE.DELETE_ERRORS_COUNT, len(fallback_failures), type=self.TYPE)
        return fallback_failures

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        written_to_cache = False
        try:
            self._cache.commit_write(digest, write_session)
            written_to_cache = True
        except Exception:
            LOGGER.warning(
                "Failed to commit write of digest to cache storage.", tags=dict(digest=digest), exc_info=True
            )

        if written_to_cache and self._defer_fallback_writes:
            write_session.seek(0)
            deferred_session = create_write_session(digest)
            while data := write_session.read(MAX_IN_MEMORY_BLOB_SIZE_BYTES):
                deferred_session.write(data)

            def deferred_submit() -> None:
                with deferred_session:
                    self._fallback.commit_write(digest, deferred_session)

            self._executor.submit(deferred_submit)
        else:
            self._fallback.commit_write(digest, write_session)

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        return self._fallback.missing_blobs(digests)

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        try:
            self._cache.bulk_update_blobs(blobs)
        except Exception:
            LOGGER.warning("Failed to bulk update blobs in cache storage.", exc_info=True)

        return self._fallback.bulk_update_blobs(blobs)

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        try:
            cache_blobs = self._cache.bulk_read_blobs(digests)
        except Exception:
            LOGGER.warning("Failed to bulk read blobs from cache storage.", exc_info=True)
            cache_blobs = {}

        publish_distribution_metric(METRIC.STORAGE.WITH_CACHE.CACHE_HIT_COUNT, len(cache_blobs))

        uncached_digests = [digest for digest in digests if cache_blobs.get(digest.hash, None) is None]

        publish_distribution_metric(METRIC.STORAGE.WITH_CACHE.CACHE_MISS_COUNT, len(digests) - len(cache_blobs))

        metric_cache_percent = 0.0
        if len(digests) > 0:
            metric_cache_percent = len(cache_blobs) / len(digests) * 100
        publish_distribution_metric(METRIC.STORAGE.WITH_CACHE.CACHE_HIT_PERCENT, metric_cache_percent)

        fallback_blobs = self._fallback.bulk_read_blobs(uncached_digests)
        cache_blobs.update(fallback_blobs)

        uncached_blobs = []
        for digest in uncached_digests:
            blob = fallback_blobs.get(digest.hash)
            if blob is not None:
                uncached_blobs.append((digest, blob))

        try:
            self._cache.bulk_update_blobs(uncached_blobs)
        except Exception:
            LOGGER.warning("Failed to add blobs to cache storage after bulk read.", exc_info=True)

        return cache_blobs



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextlib import contextmanager
from typing import Iterator, Optional

import grpc

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    ActionResult,
    Digest,
    GetActionResultRequest,
    UpdateActionResultRequest,
)
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2_grpc import ActionCacheStub
from buildgrid.server.client.retrier import GrpcRetrier
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.metadata import metadata_list


class ActionCacheClient:
    """Remote ActionCache service client helper.

    The :class:`ActionCacheClient` class comes with a generator factory function
    that can be used together with the `with` statement for context management::

        from buildgrid.server.client.actioncache import query

        with query(channel, instance='build') as action_cache:
            digest, action_result = action_cache.get(action_digest)
    """

    def __init__(
        self,
        channel: grpc.Channel,
        retries: int = 0,
        max_backoff: int = 64,
        should_backoff: bool = True,
    ) -> None:
        """Initializes a new :class:`ActionCacheClient` instance.

        Args:
            channel (grpc.Channel): a gRPC channel to the ActionCache endpoint.
        """
        self._grpc_retrier = GrpcRetrier(retries=retries, max_backoff=max_backoff, should_backoff=should_backoff)
        self.channel = channel
        self._stub = ActionCacheStub(self.channel)

    def get(self, instance_name: Optional[str], action_digest: Digest) -> Optional[ActionResult]:
        """Attempt to retrieve cached :obj:`ActionResult` for : given obj:`Action`."""
        try:
            return self._grpc_retrier.retry(self._get, instance_name, action_digest)
        except NotFoundError:
            return None

    def update(
        self, instance_name: Optional[str], action_digest: Digest, action_result: ActionResult
    ) -> Optional[ActionResult]:
        """Attempt to map in cache an :obj:`Action` to an :obj:`ActionResult`."""
        try:
            return self._grpc_retrier.retry(self._update, instance_name, action_digest, action_result)
        except NotFoundError:
            return None

    def _get(self, instance_name: Optional[str], action_digest: Digest) -> ActionResult:
        """Retrieves the cached :obj:`ActionResult` for a given :obj:`Action`.

        Args:
            instance_name (str): the instance name of the action cache.
            action_digest (:obj:`Digest`): the action's digest to query.

        Returns:
            :obj:`ActionResult`: the cached result or None if not found.

        Raises:
            grpc.RpcError: on any network or remote service error.
        """

        request = GetActionResultRequest()
        request.instance_name = instance_name or ""
        request.action_digest.CopyFrom(action_digest)

        res = self._stub.GetActionResult(request, metadata=metadata_list())
        return res

    def _update(
        self, instance_name: Optional[str], action_digest: Digest, action_result: ActionResult
    ) -> ActionResult:
        """Maps in cache an :obj:`Action` to an :obj:`ActionResult`.

        Args:
            instance_name (str): the instance name of the action cache.
            action_digest (:obj:`Digest`): the action's digest to update.
            action_result (:obj:`ActionResult`): the action's result.

        Returns:
            :obj:`ActionResult`: the cached result or None on failure.

        Raises:
            grpc.RpcError: on any network or remote service error.
        """

        request = UpdateActionResultRequest()
        request.instance_name = instance_name or ""
        request.action_digest.CopyFrom(action_digest)
        request.action_result.CopyFrom(action_result)

        res = self._stub.UpdateActionResult(request, metadata=metadata_list())
        return res

    def close(self) -> None:
        """Closes the underlying connection stubs."""
        self.channel.close()


@contextmanager
def query(
    channel: grpc.Channel,
    retries: int = 0,
    max_backoff: int = 64,
    should_backoff: bool = True,
) -> Iterator[ActionCacheClient]:
    """Context manager generator for the :class:`ActionCacheClient` class."""
    client = ActionCacheClient(channel, retries=retries, max_backoff=max_backoff, should_backoff=should_backoff)
    try:
        yield client
    finally:
        client.close()



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from buildgrid.server.exceptions import ServerError


class AuthError(ServerError):
    pass


class MissingTokenError(AuthError):
    def __init__(self) -> None:
        super().__init__("Missing authentication header field")


class InvalidAuthorizationHeaderError(AuthError):
    def __init__(self) -> None:
        super().__init__("Invalid authentication header field")


class InvalidTokenError(AuthError):
    def __init__(self) -> None:
        super().__init__("Invalid authentication token")


class SigningKeyNotFoundError(AuthError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Cannot find signing key for token: {message}")


class ExpiredTokenError(AuthError):
    def __init__(self) -> None:
        super().__init__("Expired authentication token")


class UnboundedTokenError(AuthError):
    def __init__(self, context: str = "") -> None:
        message = "Unbounded authentication token"
        if context:
            message = f"{message}: {context}"
        super().__init__(message)


class UnexpectedTokenParsingError(AuthError):
    def __init__(self) -> None:
        super().__init__("Unexpected error parsing authentication token")



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
SizeDifferentiatedStorage
=========================

A storage provider which passes requests to other storage providers
based on the size of the blob being requested.

"""


from collections import defaultdict
from contextlib import ExitStack
from typing import IO, Callable, Dict, Iterable, Iterator, List, NamedTuple, Optional, Tuple, TypedDict, TypeVar

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric
from buildgrid.server.threading import ContextThreadPoolExecutor

from .storage_abc import StorageABC

LOGGER = buildgrid_logger(__name__)


class SizeLimitedStorageType(TypedDict):
    max_size: int
    storage: StorageABC


# NOTE: This exists separately to the TypedDict to allow attribute-based access
# at runtime, rather than relying on string-based access to dictionary keys.
class _SizeLimitedStorage(NamedTuple):
    max_size: int
    storage: StorageABC


_T = TypeVar("_T")
_R = TypeVar("_R")


# wrapper functions for the bulk StorageABC interfaces
def _bulk_delete_for_storage(storage_digests: Tuple[StorageABC, List[Digest]]) -> List[str]:
    storage, digests = storage_digests
    return storage.bulk_delete(digests)


def _fmb_for_storage(storage_digests: Tuple[StorageABC, List[Digest]]) -> List[Digest]:
    storage, digests = storage_digests
    return storage.missing_blobs(digests)


def _bulk_update_for_storage(
    storage_digests: Tuple[StorageABC, List[Tuple[Digest, bytes]]]
) -> Tuple[StorageABC, List[Status]]:
    storage, digest_tuples = storage_digests
    return storage, storage.bulk_update_blobs(digest_tuples)


def _bulk_read_for_storage(storage_digests: Tuple[StorageABC, List[Digest]]) -> Dict[str, bytes]:
    storage, digests = storage_digests
    return storage.bulk_read_blobs(digests)


class SizeDifferentiatedStorage(StorageABC):
    TYPE = "SizeDifferentiated"

    def __init__(
        self, storages: List[SizeLimitedStorageType], fallback: StorageABC, thread_pool_size: Optional[int] = None
    ) -> None:
        self._stack = ExitStack()
        self._fallback_storage = fallback
        self._storages = [_SizeLimitedStorage(**storage) for storage in storages]
        self._storages.sort(key=lambda storage: storage.max_size)
        self._threadpool: Optional[ContextThreadPoolExecutor] = None
        if thread_pool_size:
            self._threadpool = ContextThreadPoolExecutor(thread_pool_size, "size-differentiated-storage")

    def _storage_from_digest(self, digest: Digest) -> StorageABC:
        for storage in self._storages:
            if digest.size_bytes < storage.max_size:
                return storage.storage
        # If the blob is too big for any of the size-limited storages,
        # put it in the fallback.
        return self._fallback_storage

    def _partition_digests(self, digests: List[Digest]) -> Dict[StorageABC, List[Digest]]:
        partition: Dict[StorageABC, List[Digest]] = defaultdict(list)
        for digest in digests:
            storage = self._storage_from_digest(digest)
            partition[storage].append(digest)
        return partition

    def _map(self, fn: Callable[[_T], _R], args: Iterable[_T]) -> Iterator[_R]:
        if self._threadpool:
            return self._threadpool.map(fn, args)
        else:
            return map(fn, args)

    def start(self) -> None:
        if self._threadpool:
            self._stack.enter_context(self._threadpool)
        for storage_tuple in self._storages:
            self._stack.enter_context(storage_tuple.storage)

    def stop(self) -> None:
        self._stack.close()
        LOGGER.info(f"Stopped {type(self).__name__}")

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        LOGGER.debug("Checking for blob.", tags=dict(digest=digest))
        storage = self._storage_from_digest(digest)
        return storage.has_blob(digest)

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        LOGGER.debug("Getting blob.", tags=dict(digest=digest))
        storage = self._storage_from_digest(digest)
        return storage.get_blob(digest)

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        LOGGER.debug("Deleting blob.", tags=dict(digest=digest))
        storage = self._storage_from_digest(digest)
        storage.delete_blob(digest)

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        LOGGER.debug("Writing blob.", tags=dict(digest=digest))
        storage = self._storage_from_digest(digest)
        storage.commit_write(digest, write_session)

    @timed(METRIC.STORAGE.BULK_DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        failed_deletions: List[str] = []
        for result in self._map(_bulk_delete_for_storage, self._partition_digests(digests).items()):
            failed_deletions.extend(result)

        publish_counter_metric(METRIC.STORAGE.DELETE_ERRORS_COUNT, len(failed_deletions), type=self.TYPE)
        return failed_deletions

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        missing_blobs: List[Digest] = []

        for result in self._map(_fmb_for_storage, self._partition_digests(digests).items()):
            missing_blobs.extend(result)

        return missing_blobs

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        partitioned_digests: Dict[StorageABC, List[Tuple[Digest, bytes]]] = defaultdict(list)
        idx_map: Dict[StorageABC, List[int]] = defaultdict(list)
        for orig_idx, digest_tuple in enumerate(blobs):
            storage = self._storage_from_digest(digest_tuple[0])
            partitioned_digests[storage].append(digest_tuple)
            idx_map[storage].append(orig_idx)

        results: List[Status] = [Status(code=code_pb2.INTERNAL, message="inconsistent batch results")] * len(blobs)
        for storage, statuses in self._map(_bulk_update_for_storage, partitioned_digests.items()):
            for status_idx, status in enumerate(statuses):
                results[idx_map[storage][status_idx]] = status
        return results

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        bulk_read_results: Dict[str, bytes] = {}
        for result in self._map(_bulk_read_for_storage, self._partition_digests(digests).items()):
            bulk_read_results.update(result)

        return bulk_read_results



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
BotsService
=================

"""

from typing import Dict, cast

import grpc
from google.protobuf import empty_pb2

from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2 import DESCRIPTOR as BOTS_DESCRIPTOR
from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2 import (
    BotSession,
    CreateBotSessionRequest,
    PostBotEventTempRequest,
    UpdateBotSessionRequest,
)
from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2_grpc import (
    BotsServicer,
    add_BotsServicer_to_server,
)
from buildgrid.server.bots.instance import BotsInterface
from buildgrid.server.decorators import rpc
from buildgrid.server.enums import BotStatus
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.servicer import InstancedServicer
from buildgrid.server.utils.cancellation import CancellationContext


def _instance_name_from_bot_name(name: str) -> str:
    names = name.split("/")
    return "/".join(names[:-1])


class BotsService(BotsServicer, InstancedServicer[BotsInterface]):
    SERVICE_NAME = "Bots"
    REGISTER_METHOD = add_BotsServicer_to_server
    FULL_NAME = BOTS_DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @rpc(instance_getter=lambda r: cast(str, r.parent))
    def CreateBotSession(self, request: CreateBotSessionRequest, context: grpc.ServicerContext) -> BotSession:
        return self.current_instance.create_bot_session(
            request.bot_session, CancellationContext(context), deadline=context.time_remaining()
        )

    @rpc(instance_getter=lambda r: _instance_name_from_bot_name(r.name))
    def UpdateBotSession(self, request: UpdateBotSessionRequest, context: grpc.ServicerContext) -> BotSession:
        if request.name != request.bot_session.name:
            raise InvalidArgumentError(
                "Name in UpdateBotSessionRequest does not match BotSession. "
                f" UpdateBotSessionRequest.name=[{request.name}] BotSession.name=[{request.bot_session.name}]"
            )

        bot_session, metadata = self.current_instance.update_bot_session(
            request.bot_session, CancellationContext(context), deadline=context.time_remaining()
        )
        context.set_trailing_metadata(metadata)  # type: ignore[arg-type]  # tricky covariance issue.

        return bot_session

    @rpc(instance_getter=lambda r: _instance_name_from_bot_name(r.name))
    def PostBotEventTemp(self, request: PostBotEventTempRequest, context: grpc.ServicerContext) -> empty_pb2.Empty:
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        return empty_pb2.Empty()

    def query_connected_bots_for_instance(self, instance_name: str) -> int:
        if instance := self.instances.get(instance_name):
            return instance.scheduler.job_assigner.listener_count(instance_name)
        return 0

    def count_bots_by_status(self, instance_name: str) -> Dict[BotStatus, int]:
        if instance := self.instances.get(instance_name):
            return instance.count_bots_by_status()
        return {}



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List

from lark import Lark
from lark.exceptions import LarkError

from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.operations.filtering import OperationFilter
from buildgrid.server.operations.filtering.interpreter import FilterTreeInterpreter


class FilterParser:
    """
    Utility class primarily used to parse a filter string and return a list a OperationFilters.
    """

    lark_parser = Lark.open_from_package(
        "buildgrid.server.operations.filtering", "filter_grammar.lark", start="filter_phrase"
    )

    @staticmethod
    def parse_listoperations_filters(filter_string: str) -> List[OperationFilter]:
        """Separate the lowercase filter string into individual components, and return a map containing
        the relevant filters to use.

        Filter strings take the following form: key1=value1&key2=value2,value3&key3=value4
        All spaces in the input are ignored.

        Filter options are separated with ampersands (&).
        If a key is repeated, the combined set of values are treated as a logical conjunction (AND).

        """
        operation_filters: List[OperationFilter] = []

        # Handle empty input separately to simplify grammar
        if not filter_string.strip():
            return operation_filters

        try:
            filter_phrase_tree = FilterParser.lark_parser.parse(filter_string)
        except LarkError as e:
            raise InvalidArgumentError(
                f"Error parsing filter string. See docs for correct filter string syntax. "
                f"Parser error follows: [{str(e)}]"
            )

        try:
            operation_filters = FilterTreeInterpreter().visit(filter_phrase_tree)
        except InvalidArgumentError as e:
            raise InvalidArgumentError(
                f"Invalid filter string [{filter_string}]. See docs for correct filter string syntax and values."
            ) from e

        return operation_filters



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from abc import ABC
from datetime import datetime
from typing import Any, List

import dateutil.parser
from dateutil.tz import tzutc

import buildgrid.server.enums as enums
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.operations.filtering import SortKey


class ValueSanitizer(ABC):
    """Base sanitizer class."""

    @property
    def valid_values(self) -> List[str]:
        """Return a list of valid values for the sanitizer.

        This is only useful for sanitizers with a finite list of valid values.

        Raises NotImplementedError for sanitizers which aren't based on an
        enum of valid values.

        """
        raise NotImplementedError()

    # TODO probably make this generic?
    def sanitize(self, value_string: str) -> Any:
        """Method that takes an input string, validates that input string,
        and transforms it to a value of another type if necessary.

        Raises InvalidArgumentError if the sanitization fails. Returns a
        value of an arbitrary type if it succeeds."""
        raise NotImplementedError()


class RegexValueSanitizer(ValueSanitizer):
    """Sanitizer for regexable patterns."""

    def __init__(self, filter_name: str, regex_pattern: str) -> None:
        self.filter_name = filter_name
        self.regex = re.compile(f"^({regex_pattern})$")

    def sanitize(self, value_string: str) -> str:
        if not self.regex.fullmatch(value_string):
            raise InvalidArgumentError(f"[{value_string}] is not a valid value for {self.filter_name}.")
        return value_string


class DatetimeValueSanitizer(ValueSanitizer):
    """Sanitizer for ISO 8601 datetime strings."""

    def __init__(self, filter_name: str) -> None:
        self.filter_name = filter_name

    def sanitize(self, value_string: str) -> datetime:
        try:
            dt = dateutil.parser.isoparse(value_string)
            # Convert to UTC and remove timezone awareness
            if dt.tzinfo:
                dt = dt.astimezone(tz=tzutc()).replace(tzinfo=None)
            return dt
        except ValueError:
            raise InvalidArgumentError(f"[{value_string}] is not a valid value for {self.filter_name}.")


class OperationStageValueSanitizer(ValueSanitizer):
    """Sanitizer for the OperationStage type.

    Matches valid OperationStage values and converts to the
    numeric representation of that stage."""

    def __init__(self, filter_name: str) -> None:
        self.filter_name = filter_name

    @property
    def valid_values(self) -> List[str]:
        return [stage.name for stage in enums.OperationStage]

    def sanitize(self, value_string: str) -> int:
        try:
            stage = value_string.upper()
            return enums.OperationStage[stage].value
        except KeyError:
            raise InvalidArgumentError(f"[{value_string}] is not a valid value for {self.filter_name}.")


class SortKeyValueSanitizer(ValueSanitizer):
    """Sanitizer for sort orders.

    Produces a SortKey tuple, which specifies both a key name and a boolean
    indicating ascending/descending order."""

    def __init__(self, filter_name: str) -> None:
        self.filter_name = filter_name

    def sanitize(self, value_string: str) -> SortKey:
        desc_key = "(desc)"
        asc_key = "(asc)"

        key_name = value_string.lower().strip()
        if not key_name:
            raise InvalidArgumentError(f"Invalid sort key [{value_string}].")
        descending = False

        if value_string.endswith(desc_key):
            descending = True
            key_name = key_name[: -len(desc_key)].strip()
            if not key_name:
                raise InvalidArgumentError(f"Invalid sort key [{value_string}].")

        elif value_string.endswith(asc_key):
            key_name = key_name[: -len(asc_key)].strip()
            if not key_name:
                raise InvalidArgumentError(f"Invalid sort key [{value_string}].")

        return SortKey(name=key_name, descending=descending)



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
RemoteStorage
==================

Forwwards storage requests to a remote storage.
"""

import io
import logging
from tempfile import NamedTemporaryFile
from typing import IO, Any, Dict, List, Optional, Sequence, Tuple

import grpc

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2, remote_execution_pb2_grpc
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2, status_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.client.authentication import ClientCredentials
from buildgrid.server.client.cas import download, upload
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.context import current_instance
from buildgrid.server.decorators import timed
from buildgrid.server.exceptions import GrpcUninitializedError, NotFoundError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metadata import metadata_list
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.settings import HASH, MAX_IN_MEMORY_BLOB_SIZE_BYTES

from .storage_abc import StorageABC

LOGGER = buildgrid_logger(__name__)


class RemoteStorage(StorageABC):
    TYPE = "Remote"

    def __init__(
        self,
        remote: str,
        instance_name: Optional[str] = None,
        channel_options: Optional[Sequence[Tuple[str, Any]]] = None,
        credentials: Optional[ClientCredentials] = None,
        retries: int = 0,
        max_backoff: int = 64,
        request_timeout: Optional[float] = None,
    ) -> None:
        self._remote_instance_name = instance_name
        self._remote = remote
        self._channel_options = channel_options
        if credentials is None:
            credentials = {}
        self.credentials = credentials
        self.retries = retries
        self.max_backoff = max_backoff
        self._request_timeout = request_timeout

        self._stub_cas: Optional[remote_execution_pb2_grpc.ContentAddressableStorageStub] = None
        self.channel: Optional[grpc.Channel] = None

    def start(self) -> None:
        if self.channel is None:
            self.channel, *_ = setup_channel(
                self._remote,
                auth_token=self.credentials.get("auth-token"),
                auth_token_refresh_seconds=self.credentials.get("token-refresh-seconds"),
                client_key=self.credentials.get("tls-client-key"),
                client_cert=self.credentials.get("tls-client-cert"),
                server_cert=self.credentials.get("tls-server-cert"),
                timeout=self._request_timeout,
            )

        if self._stub_cas is None:
            self._stub_cas = remote_execution_pb2_grpc.ContentAddressableStorageStub(self.channel)

    def stop(self) -> None:
        if self.channel is not None:
            self.channel.close()

    @property
    def remote_instance_name(self) -> str:
        if self._remote_instance_name is not None:
            return self._remote_instance_name
        return current_instance()

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        LOGGER.debug("Checking for blob.", tags=dict(digest=digest))
        if not self.missing_blobs([digest]):
            return True
        return False

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        if self.channel is None:
            raise GrpcUninitializedError("Remote CAS backend used before gRPC initialization.")

        LOGGER.debug("Getting blob.", tags=dict(digest=digest))
        with download(
            self.channel, instance=self.remote_instance_name, retries=self.retries, max_backoff=self.max_backoff
        ) as downloader:
            if digest.size_bytes > MAX_IN_MEMORY_BLOB_SIZE_BYTES:
                # Avoid storing the large blob completely in memory.
                temp_file = NamedTemporaryFile(delete=True)
                try:
                    downloader.download_file(digest, temp_file.name, queue=False)
                except NotFoundError:
                    return None
                # TODO Fix this: incompatible type "_TemporaryFileWrapper[bytes]"; expected "RawIOBase"
                reader = io.BufferedReader(temp_file)  # type: ignore[arg-type]
                reader.seek(0)
                return reader
            else:
                blob = downloader.get_blob(digest)
                if blob is not None:
                    return io.BytesIO(blob)
                else:
                    return None

    def delete_blob(self, digest: Digest) -> None:
        """The REAPI doesn't have a deletion method, so we can't support
        deletion for remote storage.
        """
        raise NotImplementedError("Deletion is not supported for remote storage!")

    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        """The REAPI doesn't have a deletion method, so we can't support
        bulk deletion for remote storage.
        """
        raise NotImplementedError("Bulk deletion is not supported for remote storage!")

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        if self.channel is None:
            raise GrpcUninitializedError("Remote CAS backend used before gRPC initialization.")

        write_session.seek(0)
        LOGGER.debug("Writing blob.", tags=dict(digest=digest))
        with upload(
            self.channel, instance=self.remote_instance_name, retries=self.retries, max_backoff=self.max_backoff
        ) as uploader:
            uploader.put_blob(write_session, digest=digest)

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        if self._stub_cas is None:
            raise GrpcUninitializedError("Remote CAS backend used before gRPC initialization.")

        # Avoid expensive string creation.
        if LOGGER.is_enabled_for(logging.DEBUG):
            if len(digests) > 100:
                LOGGER.debug(f"Missing blobs request for: {digests[:100]} (truncated)")
            else:
                LOGGER.debug(f"Missing blobs request for: {digests}")

        request = remote_execution_pb2.FindMissingBlobsRequest(instance_name=self.remote_instance_name)

        for blob in digests:
            request_digest = request.blob_digests.add()
            request_digest.hash = blob.hash
            request_digest.size_bytes = blob.size_bytes

        response = self._stub_cas.FindMissingBlobs(request, metadata=metadata_list())

        return list(response.missing_blob_digests)

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        if self._stub_cas is None or self.channel is None:
            raise GrpcUninitializedError("Remote CAS backend used before gRPC initialization.")

        sent_digests = []
        with upload(
            self.channel, instance=self.remote_instance_name, retries=self.retries, max_backoff=self.max_backoff
        ) as uploader:
            for digest, blob in blobs:
                if len(blob) != digest.size_bytes or HASH(blob).hexdigest() != digest.hash:
                    sent_digests.append(remote_execution_pb2.Digest())
                else:
                    sent_digests.append(uploader.put_blob(io.BytesIO(blob), digest=digest, queue=True))

        assert len(sent_digests) == len(blobs)

        return [
            status_pb2.Status(code=code_pb2.OK) if d.ByteSize() > 0 else status_pb2.Status(code=code_pb2.UNKNOWN)
            for d in sent_digests
        ]

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        if self._stub_cas is None or self.channel is None:
            raise GrpcUninitializedError("Remote CAS backend used before gRPC initialization.")

        # Avoid expensive string creation.
        if LOGGER.is_enabled_for(logging.DEBUG):
            LOGGER.debug(f"Bulk read blobs request for: {digests}")

        with download(
            self.channel, instance=self.remote_instance_name, retries=self.retries, max_backoff=self.max_backoff
        ) as downloader:
            results = downloader.get_available_blobs(digests)
            # Transform List of (data, digest) pairs to expected hash-blob map
            return {digest.hash: data for data, digest in results}



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Execute command
=================

Request work to be executed and monitor status of jobs.
"""

import os
import signal
import stat
import sys
from typing import Any, List, Tuple

import click
from grpc import RpcError

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2, remote_execution_pb2_grpc
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid.server.client.cas import download, upload
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.utils.digests import create_digest

from ..cli import pass_context
from .rpc_utils import cancel_operation


@click.group(name="execute", short_help="Execute simple operations.")
@click.option(
    "--remote",
    type=click.STRING,
    default="http://localhost:50051",
    show_default=True,
    help="Remote execution server's URL (port defaults to 50051 if no specified).",
)
@click.option(
    "--auth-token",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Authorization token for the remote.",
)
@click.option(
    "--client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private client key for TLS (PEM-encoded).",
)
@click.option(
    "--client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public server certificate for TLS (PEM-encoded).",
)
@click.option(
    "--instance-name", type=click.STRING, default=None, show_default=True, help="Targeted farm instance name."
)
@click.option(
    "--remote-cas",
    type=click.STRING,
    default=None,
    show_default=False,
    help="Remote CAS server's URL (defaults to --remote if not specified).",
)
@click.option(
    "--cas-client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private CAS client key for TLS (PEM-encoded).",
)
@click.option(
    "--cas-client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public CAS client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--cas-server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public CAS server certificate for TLS (PEM-encoded).",
)
@click.option("--action-id", type=str, help="Action ID.")
@click.option("--invocation-id", type=str, help="Tool invocation ID.")
@click.option("--correlation-id", type=str, help="Correlated invocation ID.")
@pass_context
def cli(
    context: Any,
    remote: str,
    instance_name: str,
    auth_token: str,
    client_key: str,
    client_cert: str,
    server_cert: str,
    remote_cas: str,
    cas_client_key: str,
    cas_client_cert: str,
    cas_server_cert: str,
    action_id: str,
    invocation_id: str,
    correlation_id: str,
) -> None:
    """Entry point for the bgd-execute CLI command group."""
    try:
        context.channel, details = setup_channel(
            remote,
            auth_token=auth_token,
            client_key=client_key,
            client_cert=client_cert,
            server_cert=server_cert,
            action_id=action_id,
            tool_invocation_id=invocation_id,
            correlated_invocations_id=correlation_id,
        )

        if remote_cas and remote_cas != remote:
            context.cas_channel, details = setup_channel(
                remote_cas,
                auth_token=auth_token,
                server_cert=cas_server_cert,
                client_key=cas_client_key,
                client_cert=cas_client_cert,
                action_id=action_id,
                tool_invocation_id=invocation_id,
                correlated_invocations_id=correlation_id,
            )
            context.remote_cas_url = remote_cas

        else:
            context.cas_channel = context.channel
            context.remote_cas_url = remote

        context.cas_client_key, context.cas_client_cert, context.cas_server_cert = details

    except InvalidArgumentError as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(-1)

    context.instance_name = instance_name


@cli.command("request-dummy", short_help="Send a dummy action.")
@click.option("--number", type=click.INT, default=1, show_default=True, help="Number of request to send.")
@click.option("--wait-for-completion", is_flag=True, help="Stream updates until jobs are completed.")
@pass_context
def request_dummy(context: Any, number: int, wait_for_completion: bool) -> None:
    click.echo("Sending execution request...")
    command = remote_execution_pb2.Command()
    command_digest = create_digest(command.SerializeToString())

    action = remote_execution_pb2.Action(command_digest=command_digest, do_not_cache=True)
    action_digest = create_digest(action.SerializeToString())

    stub = remote_execution_pb2_grpc.ExecutionStub(context.channel)

    request = remote_execution_pb2.ExecuteRequest(
        instance_name=context.instance_name, action_digest=action_digest, skip_cache_lookup=True
    )

    responses = []
    for _ in range(0, number):
        responses.append(stub.Execute(request))

    for response in responses:
        try:
            if wait_for_completion:
                result = None
                for stream in response:
                    result = stream
                    click.echo(result)

                if not result or not result.done:
                    click.echo("Result did not return True." + "Was the action uploaded to CAS?", err=True)
                    sys.exit(-1)
            else:
                # Bug in proto stubs: No overload variant of "next" matches argument type "CallIterator[Operation]"
                click.echo(next(response))  # type: ignore[call-overload]
        except RpcError as e:
            click.echo(f"Error: Requesting dummy: {e.details()}", err=True)


@cli.command("command", short_help="Send a command to be executed.")
@click.option("--output-file", nargs=1, type=click.STRING, multiple=True, help="Expected output file.")
@click.option(
    "--output-directory", default="testing", show_default=True, help="Output directory for the output files."
)
@click.option(
    "-p",
    "--platform-property",
    nargs=2,
    type=(click.STRING, click.STRING),
    multiple=True,
    help="List of key-value pairs of required platform properties.",
)
@click.option(
    "--disable-cancellation",
    is_flag=True,
    help="By default, SIGINT sends a CancelOperation request. Set this flag to disable this behavior.",
)
@click.option(
    "-c",
    "--enable-action-cache",
    is_flag=True,
    help="By default, the result is not cached and the cache isn't checked for a previous result. "
    "Set this flag to enable caching.",
)
@click.argument("input-root", nargs=1, type=click.Path(), required=True)
@click.argument("commands", nargs=-1, type=click.STRING, required=True)
@click.option(
    "--priority",
    nargs=1,
    type=click.INT,
    show_default=True,
    default=0,
    help="Set priority of execution request, the value can be negative or positive. "
    "Smaller values result in higher priority.",
)
@click.option(
    "--request-workflow", nargs=1, type=click.STRING, default="", help="The workflow of execution request, e.g., build"
)
@click.option(
    "--request-actor",
    nargs=1,
    type=click.STRING,
    default="",
    help="The actor of execution request, e.g. bgd-cli",
)
@click.option(
    "--request-subject",
    type=click.STRING,
    default="",
    help="The subject of execution request, e.g. user1",
)
@pass_context
def run_command(
    context: Any,
    input_root: str,
    commands: List[str],
    output_file: List[str],
    output_directory: str,
    platform_property: List[Tuple[str, str]],
    disable_cancellation: bool,
    enable_action_cache: bool,
    priority: int,
    request_workflow: str,
    request_actor: str,
    request_subject: str,
) -> None:
    stub = remote_execution_pb2_grpc.ExecutionStub(context.channel)

    try:
        action_digest = upload_action(
            commands, context, input_root, output_file, platform_property, enable_action_cache
        )
    except Exception as e:
        click.echo(f"Error: Uploading action: {e}", err=True)
        sys.exit(-1)

    metadata = []
    if request_workflow:
        metadata.append(("x-request-workflow", request_workflow))
    if request_actor:
        metadata.append(("x-request-actor", request_actor))
    if request_subject:
        metadata.append(("x-request-subject", request_subject))

    execution_policy = remote_execution_pb2.ExecutionPolicy(priority=priority)
    request = remote_execution_pb2.ExecuteRequest(
        instance_name=context.instance_name,
        action_digest=action_digest,
        skip_cache_lookup=not enable_action_cache,
        execution_policy=execution_policy,
    )

    response = stub.Execute(request, metadata=tuple(metadata))

    operation_name = None

    # Set up a signal handler for SIGINT
    # Needs to be done here rather than as a context manager because the operation
    # name is not known a priori
    if not disable_cancellation:

        def cancel_handler(signum: int, stack: Any) -> None:
            if operation_name is not None:
                cancel_operation(context, operation_name)
            else:
                click.echo("Unable to cancel operation: operation name unknown")
            sys.exit(1)

        old_sigint_handler = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, cancel_handler)

    # Read the response until it is finished
    stream = None
    for stream in response:
        operation_name = stream.name
        click.echo(stream)

    # The last message will be the ExecuteResponse, so read it
    execute_response = remote_execution_pb2.ExecuteResponse()
    stream.response.Unpack(execute_response)
    click.echo(f"Response status code: {execute_response.status}")

    # Restore the original signal handler
    if not disable_cancellation:
        signal.signal(signal.SIGINT, old_sigint_handler)

    try:
        with download(context.cas_channel, instance=context.instance_name) as downloader:
            for output_file_response in execute_response.result.output_files:
                path = os.path.join(output_directory, output_file_response.path)

                if not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path), exist_ok=True)

                downloader.download_file(output_file_response.digest, path)
    except Exception as e:
        click.echo(f"Error: Downloading output files: {e}", err=True)
        sys.exit(-1)

    for output_file_response in execute_response.result.output_files:
        if output_file_response.is_executable:
            path = os.path.join(output_directory, output_file_response.path)
            st = os.stat(path)
            os.chmod(path, st.st_mode | stat.S_IXUSR)


def upload_action(
    commands: List[str],
    context: Any,
    input_root: str,
    output_file: List[str],
    platform_property: List[Tuple[str, str]],
    enable_action_cache: bool,
) -> Digest:
    with upload(context.cas_channel, instance=context.instance_name) as uploader:
        command = remote_execution_pb2.Command()

        for arg in commands:
            command.arguments.extend([arg])

        command.output_paths.extend(output_file)

        for attribute_name, attribute_value in platform_property:
            new_property = command.platform.properties.add()
            new_property.name = attribute_name
            new_property.value = attribute_value

        command_digest = uploader.put_message(command, queue=True)

        click.echo(f"Sent command=[{command_digest}]")

        # TODO: Check for missing blobs
        input_root_digest = uploader.upload_directory(input_root)

        click.echo(f"Sent input=[{input_root_digest}]")

        action = remote_execution_pb2.Action(
            command_digest=command_digest, input_root_digest=input_root_digest, do_not_cache=not enable_action_cache
        )
        action.platform.CopyFrom(command.platform)

        action_digest = uploader.put_message(action, queue=True)

        click.echo(f"Sent action=[{action_digest}]")

        return action_digest



import pytest

from buildgrid.server.app.settings.parser import validate_config


@pytest.mark.parametrize(
    "config_file",
    [
        "buildgrid/server/app/settings/reference.yml",
        "data/config/all-in-one.yml",
        "data/config/artifacts.yml",
        "data/config/basic-with-disk.yml",
        "data/config/bots-interface.yml",
        "data/config/cache.yml",
        "data/config/controller.yml",
        "data/config/default.yml",
        "data/config/index-sqlite-no-execution.yml",
        "data/config/index-sqlite.yml",
        "data/config/monitoring-controller.yml",
        "data/config/multi-layer-storage.yml",
        "data/config/multi-level-cache/buildgrid.yml",
        "data/config/multi-level-cache/shared-lru.yml",
        "data/config/redis-cache.yml",
        "data/config/redis-index/action-cache.yml",
        "data/config/redis-index/cas.yml",
        "data/config/redis-index/execution.yml",
        "data/config/redis-sentinel/buildgrid.yml",
        "data/config/s3-cas/action-cache.yml",
        "data/config/s3-cas/cas.yml",
        "data/config/s3-cas/execution.yml",
        "data/config/s3-indexed-cas.yml",
        "data/config/storage-redis.yml",
        "data/config/storage-s3.yml",
        "data/config/storage.yml",
        "data/config/with-metering.yml",
        "data/config/with-pgbouncer.yml",
        "docs/source/data/basic-disk-cas.yml",
        "docs/source/data/bazel-example-server.yml",
        "docs/source/data/buildstream-example-server.yml",
        "docs/source/data/cas-and-ac.yml",
        "docs/source/data/cas-example-server.yml",
        "docs/source/data/execution-and-bots.yml",
        "docs/source/data/sqlite-index-cas-only.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/bgd-cleanup.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/bgd-db-init.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/bgd-inmem.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/bgd-pges-indexedcas.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/bgd-pges.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/bgd-sqlite-indexedcas.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/bgd-sqlite.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/index-s3-inlined.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/index-s3-noninlined.yml",
        "tests/stress-testing-dockerfiles/mnt/configs/index-s3.yml",
    ],
)
def test_example_config(config_file):
    validate_config(config_file, strict=True, fail_deprecations=True)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 The Bazel Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import abc
import buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2
import grpc

class FetchStub:
    """The Fetch service resolves or fetches assets referenced by URI and
    Qualifiers, returning a Digest for the content in
    [ContentAddressableStorage][build.bazel.remote.execution.v2.ContentAddressableStorage].

    As with other services in the Remote Execution API, any call may return an
    error with a [RetryInfo][google.rpc.RetryInfo] error detail providing
    information about when the client should retry the request; clients SHOULD
    respect the information provided.
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    FetchBlob: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchBlobRequest,
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchBlobResponse,
    ]
    """Resolve or fetch referenced assets, making them available to the caller and
    other consumers in the [ContentAddressableStorage][build.bazel.remote.execution.v2.ContentAddressableStorage].

    Servers *MAY* fetch content that they do not already have cached, for any
    URLs they support.

    Servers *SHOULD* ensure that referenced files are present in the CAS at the
    time of the response, and (if supported) that they will remain available
    for a reasonable period of time. The lifetimes of the referenced blobs *SHOULD*
    be increased if necessary and applicable.
    In the event that a client receives a reference to content that is no
    longer present, it *MAY* re-issue the request with
    `oldest_content_accepted` set to a more recent timestamp than the original
    attempt, to induce a re-fetch from origin.

    Servers *MAY* cache fetched content and reuse it for subsequent requests,
    subject to `oldest_content_accepted`.

    Servers *MAY* support the complementary [Push][build.bazel.remote.asset.v1.Push]
    API and allow content to be directly inserted for use in future fetch
    responses.

    Servers *MUST* ensure Fetch'd content matches all the specified
    qualifiers except in the case of previously Push'd resources, for which
    the server *MAY* trust the pushing client to have set the qualifiers
    correctly, without validation.

    Servers not implementing the complementary [Push][build.bazel.remote.asset.v1.Push]
    API *MUST* reject requests containing qualifiers it does not support.

    Servers *MAY* transform assets as part of the fetch. For example a
    tarball fetched by [FetchDirectory][build.bazel.remote.asset.v1.Fetch.FetchDirectory]
    might be unpacked, or a Git repository
    fetched by [FetchBlob][build.bazel.remote.asset.v1.Fetch.FetchBlob]
    might be passed through `git-archive`.

    Errors handling the requested assets will be returned as gRPC Status errors
    here; errors outside the server's control will be returned inline in the
    `status` field of the response (see comment there for details).
    The possible RPC errors include:
    * `INVALID_ARGUMENT`: One or more arguments were invalid, such as a
      qualifier that is not supported by the server.
    * `RESOURCE_EXHAUSTED`: There is insufficient quota of some resource to
      perform the requested operation. The client may retry after a delay.
    * `UNAVAILABLE`: Due to a transient condition the operation could not be
      completed. The client should retry.
    * `INTERNAL`: An internal error occurred while performing the operation.
      The client should retry.
    * `DEADLINE_EXCEEDED`: The fetch could not be completed within the given
      RPC deadline. The client should retry for at least as long as the value
      provided in `timeout` field of the request.

    In the case of unsupported qualifiers, the server *SHOULD* additionally
    send a [BadRequest][google.rpc.BadRequest] error detail where, for each
    unsupported qualifier, there is a `FieldViolation` with a `field` of
    `qualifiers.name` and a `description` of `"{qualifier}" not supported`
    indicating the name of the unsupported qualifier.
    """
    FetchDirectory: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchDirectoryRequest,
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchDirectoryResponse,
    ]

class FetchServicer(metaclass=abc.ABCMeta):
    """The Fetch service resolves or fetches assets referenced by URI and
    Qualifiers, returning a Digest for the content in
    [ContentAddressableStorage][build.bazel.remote.execution.v2.ContentAddressableStorage].

    As with other services in the Remote Execution API, any call may return an
    error with a [RetryInfo][google.rpc.RetryInfo] error detail providing
    information about when the client should retry the request; clients SHOULD
    respect the information provided.
    """

    @abc.abstractmethod
    def FetchBlob(
        self,
        request: buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchBlobRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchBlobResponse:
        """Resolve or fetch referenced assets, making them available to the caller and
        other consumers in the [ContentAddressableStorage][build.bazel.remote.execution.v2.ContentAddressableStorage].

        Servers *MAY* fetch content that they do not already have cached, for any
        URLs they support.

        Servers *SHOULD* ensure that referenced files are present in the CAS at the
        time of the response, and (if supported) that they will remain available
        for a reasonable period of time. The lifetimes of the referenced blobs *SHOULD*
        be increased if necessary and applicable.
        In the event that a client receives a reference to content that is no
        longer present, it *MAY* re-issue the request with
        `oldest_content_accepted` set to a more recent timestamp than the original
        attempt, to induce a re-fetch from origin.

        Servers *MAY* cache fetched content and reuse it for subsequent requests,
        subject to `oldest_content_accepted`.

        Servers *MAY* support the complementary [Push][build.bazel.remote.asset.v1.Push]
        API and allow content to be directly inserted for use in future fetch
        responses.

        Servers *MUST* ensure Fetch'd content matches all the specified
        qualifiers except in the case of previously Push'd resources, for which
        the server *MAY* trust the pushing client to have set the qualifiers
        correctly, without validation.

        Servers not implementing the complementary [Push][build.bazel.remote.asset.v1.Push]
        API *MUST* reject requests containing qualifiers it does not support.

        Servers *MAY* transform assets as part of the fetch. For example a
        tarball fetched by [FetchDirectory][build.bazel.remote.asset.v1.Fetch.FetchDirectory]
        might be unpacked, or a Git repository
        fetched by [FetchBlob][build.bazel.remote.asset.v1.Fetch.FetchBlob]
        might be passed through `git-archive`.

        Errors handling the requested assets will be returned as gRPC Status errors
        here; errors outside the server's control will be returned inline in the
        `status` field of the response (see comment there for details).
        The possible RPC errors include:
        * `INVALID_ARGUMENT`: One or more arguments were invalid, such as a
          qualifier that is not supported by the server.
        * `RESOURCE_EXHAUSTED`: There is insufficient quota of some resource to
          perform the requested operation. The client may retry after a delay.
        * `UNAVAILABLE`: Due to a transient condition the operation could not be
          completed. The client should retry.
        * `INTERNAL`: An internal error occurred while performing the operation.
          The client should retry.
        * `DEADLINE_EXCEEDED`: The fetch could not be completed within the given
          RPC deadline. The client should retry for at least as long as the value
          provided in `timeout` field of the request.

        In the case of unsupported qualifiers, the server *SHOULD* additionally
        send a [BadRequest][google.rpc.BadRequest] error detail where, for each
        unsupported qualifier, there is a `FieldViolation` with a `field` of
        `qualifiers.name` and a `description` of `"{qualifier}" not supported`
        indicating the name of the unsupported qualifier.
        """
    @abc.abstractmethod
    def FetchDirectory(
        self,
        request: buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchDirectoryRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.FetchDirectoryResponse: ...

def add_FetchServicer_to_server(servicer: FetchServicer, server: grpc.Server) -> None: ...

class PushStub:
    """The Push service is complementary to the Fetch, and allows for
    associating contents of URLs to be returned in future Fetch API calls.

    As with other services in the Remote Execution API, any call may return an
    error with a [RetryInfo][google.rpc.RetryInfo] error detail providing
    information about when the client should retry the request; clients SHOULD
    respect the information provided.
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    PushBlob: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushBlobRequest,
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushBlobResponse,
    ]
    """These APIs associate the identifying information of a resource, as
    indicated by URI and optionally Qualifiers, with content available in the
    CAS. For example, associating a repository url and a commit id with a
    Directory Digest.

    Servers *SHOULD* only allow trusted clients to associate content, and *MAY*
    only allow certain URIs to be pushed.

    Clients *MUST* ensure associated content is available in CAS prior to
    pushing.

    Clients *MUST* ensure the Qualifiers listed correctly match the contents,
    and Servers *MAY* trust these values without validation.
    Fetch servers *MAY* require exact match of all qualifiers when returning
    content previously pushed, or allow fetching content with only a subset of
    the qualifiers specified on Push.

    Clients can specify expiration information that the server *SHOULD*
    respect. Subsequent requests can be used to alter the expiration time.

    A minimal compliant Fetch implementation may support only Push'd content
    and return `NOT_FOUND` for any resource that was not pushed first.
    Alternatively, a compliant implementation may choose to not support Push
    and only return resources that can be Fetch'd from origin.

    Errors will be returned as gRPC Status errors.
    The possible RPC errors include:
    * `INVALID_ARGUMENT`: One or more arguments to the RPC were invalid.
    * `RESOURCE_EXHAUSTED`: There is insufficient quota of some resource to
      perform the requested operation. The client may retry after a delay.
    * `UNAVAILABLE`: Due to a transient condition the operation could not be
      completed. The client should retry.
    * `INTERNAL`: An internal error occurred while performing the operation.
      The client should retry.
    """
    PushDirectory: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushDirectoryRequest,
        buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushDirectoryResponse,
    ]

class PushServicer(metaclass=abc.ABCMeta):
    """The Push service is complementary to the Fetch, and allows for
    associating contents of URLs to be returned in future Fetch API calls.

    As with other services in the Remote Execution API, any call may return an
    error with a [RetryInfo][google.rpc.RetryInfo] error detail providing
    information about when the client should retry the request; clients SHOULD
    respect the information provided.
    """

    @abc.abstractmethod
    def PushBlob(
        self,
        request: buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushBlobRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushBlobResponse:
        """These APIs associate the identifying information of a resource, as
        indicated by URI and optionally Qualifiers, with content available in the
        CAS. For example, associating a repository url and a commit id with a
        Directory Digest.

        Servers *SHOULD* only allow trusted clients to associate content, and *MAY*
        only allow certain URIs to be pushed.

        Clients *MUST* ensure associated content is available in CAS prior to
        pushing.

        Clients *MUST* ensure the Qualifiers listed correctly match the contents,
        and Servers *MAY* trust these values without validation.
        Fetch servers *MAY* require exact match of all qualifiers when returning
        content previously pushed, or allow fetching content with only a subset of
        the qualifiers specified on Push.

        Clients can specify expiration information that the server *SHOULD*
        respect. Subsequent requests can be used to alter the expiration time.

        A minimal compliant Fetch implementation may support only Push'd content
        and return `NOT_FOUND` for any resource that was not pushed first.
        Alternatively, a compliant implementation may choose to not support Push
        and only return resources that can be Fetch'd from origin.

        Errors will be returned as gRPC Status errors.
        The possible RPC errors include:
        * `INVALID_ARGUMENT`: One or more arguments to the RPC were invalid.
        * `RESOURCE_EXHAUSTED`: There is insufficient quota of some resource to
          perform the requested operation. The client may retry after a delay.
        * `UNAVAILABLE`: Due to a transient condition the operation could not be
          completed. The client should retry.
        * `INTERNAL`: An internal error occurred while performing the operation.
          The client should retry.
        """
    @abc.abstractmethod
    def PushDirectory(
        self,
        request: buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushDirectoryRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.build.bazel.remote.asset.v1.remote_asset_pb2.PushDirectoryResponse: ...

def add_PushServicer_to_server(servicer: PushServicer, server: grpc.Server) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2016 Google Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import abc
import collections.abc
import buildgrid._protos.google.bytestream.bytestream_pb2
import grpc

class ByteStreamStub:
    """#### Introduction

    The Byte Stream API enables a client to read and write a stream of bytes to
    and from a resource. Resources have names, and these names are supplied in
    the API calls below to identify the resource that is being read from or
    written to.

    All implementations of the Byte Stream API export the interface defined here:

    * `Read()`: Reads the contents of a resource.

    * `Write()`: Writes the contents of a resource. The client can call `Write()`
      multiple times with the same resource and can check the status of the write
      by calling `QueryWriteStatus()`.

    #### Service parameters and metadata

    The ByteStream API provides no direct way to access/modify any metadata
    associated with the resource.

    #### Errors

    The errors returned by the service are in the Google canonical error space.
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    Read: grpc.UnaryStreamMultiCallable[
        buildgrid._protos.google.bytestream.bytestream_pb2.ReadRequest,
        buildgrid._protos.google.bytestream.bytestream_pb2.ReadResponse,
    ]
    """`Read()` is used to retrieve the contents of a resource as a sequence
    of bytes. The bytes are returned in a sequence of responses, and the
    responses are delivered as the results of a server-side streaming RPC.
    """
    Write: grpc.StreamUnaryMultiCallable[
        buildgrid._protos.google.bytestream.bytestream_pb2.WriteRequest,
        buildgrid._protos.google.bytestream.bytestream_pb2.WriteResponse,
    ]
    """`Write()` is used to send the contents of a resource as a sequence of
    bytes. The bytes are sent in a sequence of request protos of a client-side
    streaming RPC.

    A `Write()` action is resumable. If there is an error or the connection is
    broken during the `Write()`, the client should check the status of the
    `Write()` by calling `QueryWriteStatus()` and continue writing from the
    returned `committed_size`. This may be less than the amount of data the
    client previously sent.

    Calling `Write()` on a resource name that was previously written and
    finalized could cause an error, depending on whether the underlying service
    allows over-writing of previously written resources.

    When the client closes the request channel, the service will respond with
    a `WriteResponse`. The service will not view the resource as `complete`
    until the client has sent a `WriteRequest` with `finish_write` set to
    `true`. Sending any requests on a stream after sending a request with
    `finish_write` set to `true` will cause an error. The client **should**
    check the `WriteResponse` it receives to determine how much data the
    service was able to commit and whether the service views the resource as
    `complete` or not.
    """
    QueryWriteStatus: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.google.bytestream.bytestream_pb2.QueryWriteStatusRequest,
        buildgrid._protos.google.bytestream.bytestream_pb2.QueryWriteStatusResponse,
    ]
    """`QueryWriteStatus()` is used to find the `committed_size` for a resource
    that is being written, which can then be used as the `write_offset` for
    the next `Write()` call.

    If the resource does not exist (i.e., the resource has been deleted, or the
    first `Write()` has not yet reached the service), this method returns the
    error `NOT_FOUND`.

    The client **may** call `QueryWriteStatus()` at any time to determine how
    much data has been processed for this resource. This is useful if the
    client is buffering data and needs to know which data can be safely
    evicted. For any sequence of `QueryWriteStatus()` calls for a given
    resource name, the sequence of returned `committed_size` values will be
    non-decreasing.
    """

class ByteStreamServicer(metaclass=abc.ABCMeta):
    """#### Introduction

    The Byte Stream API enables a client to read and write a stream of bytes to
    and from a resource. Resources have names, and these names are supplied in
    the API calls below to identify the resource that is being read from or
    written to.

    All implementations of the Byte Stream API export the interface defined here:

    * `Read()`: Reads the contents of a resource.

    * `Write()`: Writes the contents of a resource. The client can call `Write()`
      multiple times with the same resource and can check the status of the write
      by calling `QueryWriteStatus()`.

    #### Service parameters and metadata

    The ByteStream API provides no direct way to access/modify any metadata
    associated with the resource.

    #### Errors

    The errors returned by the service are in the Google canonical error space.
    """

    @abc.abstractmethod
    def Read(
        self,
        request: buildgrid._protos.google.bytestream.bytestream_pb2.ReadRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[buildgrid._protos.google.bytestream.bytestream_pb2.ReadResponse]:
        """`Read()` is used to retrieve the contents of a resource as a sequence
        of bytes. The bytes are returned in a sequence of responses, and the
        responses are delivered as the results of a server-side streaming RPC.
        """
    @abc.abstractmethod
    def Write(
        self,
        request_iterator: collections.abc.Iterator[buildgrid._protos.google.bytestream.bytestream_pb2.WriteRequest],
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.google.bytestream.bytestream_pb2.WriteResponse:
        """`Write()` is used to send the contents of a resource as a sequence of
        bytes. The bytes are sent in a sequence of request protos of a client-side
        streaming RPC.

        A `Write()` action is resumable. If there is an error or the connection is
        broken during the `Write()`, the client should check the status of the
        `Write()` by calling `QueryWriteStatus()` and continue writing from the
        returned `committed_size`. This may be less than the amount of data the
        client previously sent.

        Calling `Write()` on a resource name that was previously written and
        finalized could cause an error, depending on whether the underlying service
        allows over-writing of previously written resources.

        When the client closes the request channel, the service will respond with
        a `WriteResponse`. The service will not view the resource as `complete`
        until the client has sent a `WriteRequest` with `finish_write` set to
        `true`. Sending any requests on a stream after sending a request with
        `finish_write` set to `true` will cause an error. The client **should**
        check the `WriteResponse` it receives to determine how much data the
        service was able to commit and whether the service views the resource as
        `complete` or not.
        """
    @abc.abstractmethod
    def QueryWriteStatus(
        self,
        request: buildgrid._protos.google.bytestream.bytestream_pb2.QueryWriteStatusRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.google.bytestream.bytestream_pb2.QueryWriteStatusResponse:
        """`QueryWriteStatus()` is used to find the `committed_size` for a resource
        that is being written, which can then be used as the `write_offset` for
        the next `Write()` call.

        If the resource does not exist (i.e., the resource has been deleted, or the
        first `Write()` has not yet reached the service), this method returns the
        error `NOT_FOUND`.

        The client **may** call `QueryWriteStatus()` at any time to determine how
        much data has been processed for this resource. This is useful if the
        client is buffering data and needs to know which data can be safely
        evicted. For any sequence of `QueryWriteStatus()` calls for a given
        resource name, the sequence of returned `committed_size` values will be
        non-decreasing.
        """

def add_ByteStreamServicer_to_server(servicer: ByteStreamServicer, server: grpc.Server) -> None: ...



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
SQL Storage
==================

A CAS storage which stores blobs in a SQL database

"""

import itertools
from io import BytesIO
from typing import IO, Dict, Iterator, List, Optional, Sequence, Tuple, TypedDict, cast

from sqlalchemy import delete, func, select
from sqlalchemy.dialects.postgresql import insert as postgresql_insert
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm.exc import StaleDataError

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.cas.storage.storage_abc import StorageABC
from buildgrid.server.decorators import timed
from buildgrid.server.exceptions import StorageFullError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.sql.models import BlobEntry
from buildgrid.server.sql.provider import SqlProvider
from buildgrid.server.utils.digests import validate_digest_data

LOGGER = buildgrid_logger(__name__)


class DigestRow(TypedDict):
    digest_hash: str
    digest_size_bytes: int
    data: bytes


class SQLStorage(StorageABC):
    TYPE = "SQL"

    def __init__(self, sql_provider: SqlProvider, *, sql_ro_provider: Optional[SqlProvider] = None) -> None:
        self._sql = sql_provider
        self._sql_ro = sql_ro_provider or sql_provider
        self._inclause_limit = self._sql.default_inlimit

        supported_dialects = ["postgresql", "sqlite"]

        if self._sql.dialect not in supported_dialects:
            raise RuntimeError(
                f"Unsupported dialect {self._sql.dialect}."
                f"SQLStorage only supports the following dialects: {supported_dialects}"
            )

        # Make a test query against the database to ensure the connection is valid
        with self._sql.session() as session:
            session.query(BlobEntry).first()

    def _sqlite_bulk_insert(self, new_rows: List[DigestRow]) -> None:
        with self._sql.session() as session:
            session.execute(sqlite_insert(BlobEntry).values(new_rows).on_conflict_do_nothing())

    def _postgresql_bulk_insert(self, new_rows: List[DigestRow]) -> None:
        with self._sql.session() as session:
            session.execute(postgresql_insert(BlobEntry).values(new_rows).on_conflict_do_nothing())

    def _bulk_insert(self, digests: List[Tuple[Digest, bytes]]) -> None:
        # Sort digests by hash to ensure consistent order to minimize deadlocks
        # when BatchUpdateBlobs requests have overlapping blobs
        new_rows: List[DigestRow] = [
            {"digest_hash": digest.hash, "digest_size_bytes": digest.size_bytes, "data": blob}
            for (digest, blob) in sorted(digests, key=lambda x: x[0].hash)
        ]

        if self._sql.dialect == "sqlite":
            self._sqlite_bulk_insert(new_rows)
        elif self._sql.dialect == "postgresql":
            self._postgresql_bulk_insert(new_rows)
        else:
            raise RuntimeError(f"Unsupported dialect {self._sql.dialect} for bulk_insert")

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        statement = select(func.count(BlobEntry.digest_hash)).where(BlobEntry.digest_hash == digest.hash)
        with self._sql_ro.session() as session:
            return session.execute(statement).scalar() == 1

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        statement = select(BlobEntry.data).where(BlobEntry.digest_hash == digest.hash)
        with self._sql_ro.session() as session:
            result = session.execute(statement).scalar()
        if result is not None:
            return BytesIO(result)
        return None

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        statement = delete(BlobEntry).where(BlobEntry.digest_hash == digest.hash)
        with self._sql.session() as session:
            # Set synchronize_session to false as we don't have any local session objects
            # to keep in sync
            session.execute(statement, execution_options={"synchronize_session": False})

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        write_session.seek(0)
        blob = write_session.read()
        try:
            self._bulk_insert([(digest, blob)])
        except DBAPIError as error:
            # Error has pgcode attribute (Postgres only)
            if hasattr(error.orig, "pgcode"):
                # imported here to avoid global dependency on psycopg2
                from psycopg2.errors import DiskFull, Error, OutOfMemory

                # 53100 == DiskFull && 53200 == OutOfMemory
                original_exception = cast(Error, error.orig)
                if isinstance(original_exception, (DiskFull, OutOfMemory)):
                    raise StorageFullError(
                        f"Postgres Error: {original_exception.pgerror} ({original_exception.pgcode}"
                    ) from error
            raise

    def _partitioned_hashes(self, digests: Sequence[Digest]) -> Iterator[Iterator[str]]:
        """Given a long list of digests, split it into parts no larger than
        _inclause_limit and yield the hashes in each part.
        """
        for part_start in range(0, len(digests), self._inclause_limit):
            part_end = min(len(digests), part_start + self._inclause_limit)
            part_digests = itertools.islice(digests, part_start, part_end)
            yield (digest.hash for digest in part_digests)

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        found_hashes = set()
        with self._sql_ro.session() as session:
            for part in self._partitioned_hashes(digests):
                stmt = select(BlobEntry.digest_hash).where(BlobEntry.digest_hash.in_(part))
                for row in session.execute(stmt):
                    found_hashes.add(row.digest_hash)

        return [digest for digest in digests if digest.hash not in found_hashes]

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(  # pylint: disable=arguments-renamed
        self, digest_blob_pairs: List[Tuple[Digest, bytes]]
    ) -> List[Status]:
        """Implement the StorageABC's bulk_update_blobs method.

        The StorageABC interface takes in a list of digest/blob pairs and
        returns a list of results. The list of results MUST be ordered to
        correspond with the order of the input list."""
        results = []

        pairs_to_insert = []
        for digest, blob in digest_blob_pairs:
            if validate_digest_data(digest, blob):
                results.append(Status(code=code_pb2.OK))
                pairs_to_insert.append((digest, blob))
            else:
                results.append(Status(code=code_pb2.INVALID_ARGUMENT, message="Data doesn't match hash"))

        self._bulk_insert(pairs_to_insert)
        return results

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        # Fetch all of the digests in the database
        results: Dict[str, bytes] = {}
        with self._sql_ro.session() as session:
            results = {
                digest_hash: data
                for part in self._partitioned_hashes(digests)
                for digest_hash, data in session.execute(
                    select(BlobEntry.digest_hash, BlobEntry.data).where(BlobEntry.digest_hash.in_(part))
                )
            }
        return results

    @timed(METRIC.STORAGE.BULK_DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        hashes = [x.hash for x in digests]

        # Make sure we don't exceed maximum size of an IN clause
        n = self._inclause_limit
        hash_chunks = [hashes[i : i + n] for i in range(0, len(hashes), n)]

        # We will not raise, rollback, or log on StaleDataErrors.
        # These errors occur when we delete fewer rows than we were expecting.
        # This is fine, since the missing rows will get deleted eventually.
        # When running bulk_deletes concurrently, StaleDataErrors
        # occur too often to log.
        num_blobs_deleted = 0
        with self._sql.session(exceptions_to_not_rollback_on=[StaleDataError]) as session:
            for chunk in hash_chunks:
                # Do not wait for locks when deleting rows. Skip locked rows to
                # avoid deadlocks.
                stmt = delete(BlobEntry).where(
                    BlobEntry.digest_hash.in_(
                        select(BlobEntry.digest_hash)
                        .where(BlobEntry.digest_hash.in_(chunk))
                        .with_for_update(skip_locked=True)
                    )
                )
                # Set synchronize_session to false as we don't have any local session objects
                # to keep in sync
                num_blobs_deleted += session.execute(stmt, execution_options={"synchronize_session": False}).rowcount
        LOGGER.info(
            "blobs deleted from storage.", tags=dict(deleted_count=num_blobs_deleted, digest_count=len(digests))
        )

        # bulk_delete is typically expected to return the digests that were not deleted,
        # but delete only returns the number of rows deleted and not what was/wasn't
        # deleted. Getting this info would require extra queries, so assume that
        # everything was either deleted or already deleted. Failures will continue to throw
        return []



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Execute command
=================

Request work to be executed and monitor status of jobs.
"""

import os
import sys
from typing import Any

import click

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2
from buildgrid.server.client.cas import download, merkle_tree_maker, upload
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.exceptions import InvalidArgumentError, NotFoundError
from buildgrid.server.utils.digests import create_digest, parse_digest

from ..cli import pass_context


@click.group(name="cas", short_help="Interact with the CAS server.")
@click.option(
    "--remote",
    type=click.STRING,
    default="http://localhost:50051",
    show_default=True,
    help="Remote execution server's URL (port defaults to 50051 if no specified).",
)
@click.option(
    "--auth-token",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Authorization token for the remote.",
)
@click.option(
    "--client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private client key for TLS (PEM-encoded).",
)
@click.option(
    "--client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public server certificate for TLS (PEM-encoded)",
)
@click.option(
    "--instance-name", type=click.STRING, default=None, show_default=True, help="Targeted farm instance name."
)
@click.option("--action-id", type=str, help="Action ID.")
@click.option("--invocation-id", type=str, help="Tool invocation ID.")
@click.option("--correlation-id", type=str, help="Correlated invocation ID.")
@pass_context
def cli(
    context: Any,
    remote: str,
    instance_name: str,
    auth_token: str,
    client_key: str,
    client_cert: str,
    server_cert: str,
    action_id: str,
    invocation_id: str,
    correlation_id: str,
) -> None:
    """Entry point for the bgd-cas CLI command group."""
    try:
        context.channel, _ = setup_channel(
            remote,
            auth_token=auth_token,
            client_key=client_key,
            client_cert=client_cert,
            server_cert=server_cert,
            action_id=action_id,
            tool_invocation_id=invocation_id,
            correlated_invocations_id=correlation_id,
        )

    except InvalidArgumentError as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(-1)

    context.instance_name = instance_name


@cli.command("upload-dummy", short_help="Upload a dummy action. Should be used with `execute dummy-request`")
@pass_context
def upload_dummy(context: Any) -> None:
    command = remote_execution_pb2.Command()
    try:
        with upload(context.channel, instance=context.instance_name) as uploader:
            command_digest = uploader.put_message(command)
    except Exception as e:
        click.echo(f"Error: Uploading dummy: {e}", err=True)
        sys.exit(-1)

    if command_digest.ByteSize():
        click.echo(f'Success: Pushed Command, digest=["{command_digest.hash}/{command_digest.size_bytes}]"')
    else:
        click.echo("Error: Failed pushing empty Command.", err=True)

    action = remote_execution_pb2.Action(command_digest=command_digest, do_not_cache=True)

    with upload(context.channel, instance=context.instance_name) as uploader:
        action_digest = uploader.put_message(action)

    if action_digest.ByteSize():
        click.echo(f'Success: Pushed Action, digest=["{action_digest.hash}/{action_digest.size_bytes}]"')
    else:
        click.echo("Error: Failed pushing empty Action.", err=True)


@cli.command("upload-file", short_help="Upload files to the CAS server.")
@click.argument("file_path", nargs=-1, type=click.Path(exists=True, dir_okay=False), required=True)
@click.option("--verify", is_flag=True, show_default=True, help="Check uploaded files integrity.")
@pass_context
def upload_file(context: Any, file_path: str, verify: bool) -> None:
    sent_digests = []
    try:
        with upload(context.channel, instance=context.instance_name) as uploader:
            for path in file_path:
                if not os.path.isabs(path):
                    path = os.path.relpath(path)

                click.echo(f"Queueing path=[{path}]")

                file_digest = uploader.upload_file(path, queue=True)

                sent_digests.append((file_digest, path))
    except Exception as e:
        click.echo(f"Error: Uploading file: {e}", err=True)
        sys.exit(-1)

    for sent_file_digest, sent_file_path in sent_digests:
        if verify and sent_file_digest.size_bytes != os.stat(sent_file_path).st_size:
            click.echo(f"Error: Failed to verify '{sent_file_path}'", err=True)
        elif sent_file_digest.ByteSize():
            click.echo(
                f"Success: Pushed path=[{sent_file_path}] with digest="
                f"[{sent_file_digest.hash}/{sent_file_digest.size_bytes}]"
            )
        else:
            click.echo(f"Error: Failed pushing path=[{sent_file_path}]", err=True)


@cli.command("upload-dir", short_help="Upload a directory to the CAS server.")
@click.argument("directory-path", nargs=1, type=click.Path(exists=True, file_okay=False), required=True)
@click.option("--verify", is_flag=True, show_default=True, help="Check uploaded directory's integrity.")
@pass_context
def upload_directory(context: Any, directory_path: str, verify: bool) -> None:
    sent_digests = []
    try:
        with upload(context.channel, instance=context.instance_name) as uploader:
            for node, blob, path in merkle_tree_maker(directory_path):
                if not os.path.isabs(directory_path):
                    path = os.path.relpath(path)
                click.echo(f"Queueing path=[{path}]")

                node_digest = uploader.put_blob(blob, digest=node.digest, queue=True)
                sent_digests.append((node_digest, path))
    except Exception as e:
        click.echo(f"Error: Uploading directory: {e}", err=True)
        sys.exit(-1)

    for node_digest, node_path in sent_digests:
        if verify and (os.path.isfile(node_path) and node_digest.size_bytes != os.stat(node_path).st_size):
            click.echo(f"Error: Failed to verify path=[{node_path}]", err=True)
        elif node_digest.ByteSize():
            click.echo(
                f"Success: Pushed path=[{node_path}] with digest=" f"[{node_digest.hash}/{node_digest.size_bytes}]"
            )
        else:
            click.echo(f"Error: Failed pushing path=[{node_path}]", err=True)


@cli.command(
    "download-file",
    short_help="Download one or more files from the CAS server. "
    "(Specified as a space-separated list of DIGEST FILE_PATH)",
)
@click.argument("digest-path-list", nargs=-1, type=str, required=True)  # 'digest path' pairs
@click.option("--verify", is_flag=True, show_default=True, help="Check downloaded file's integrity.")
@pass_context
def download_file(context: Any, digest_path_list: str, verify: bool) -> None:
    # Downloading files:
    downloaded_files = {}
    try:
        with download(context.channel, instance=context.instance_name) as downloader:
            for digest_string, file_path in zip(digest_path_list[0::2], digest_path_list[1::2]):
                if os.path.exists(file_path):
                    click.echo("Error: Invalid value for " + f"path=[{file_path}] already exists.", err=True)
                    continue

                digest = parse_digest(digest_string)
                assert digest

                downloader.download_file(digest, file_path)
                downloaded_files[file_path] = digest
    except NotFoundError:
        click.echo("Error: Blob not found in CAS", err=True)
        sys.exit(-1)
    except Exception as e:
        click.echo(f"Error: Downloading file: {e}", err=True)
        sys.exit(-1)

    # Verifying:
    for file_path, digest in downloaded_files.items():
        if verify:
            with open(file_path, "rb") as byte_file:
                file_digest = create_digest(byte_file.read())

            if file_digest != digest:
                click.echo(f"Error: Failed to verify path=[{file_path}]", err=True)
                continue

        if os.path.isfile(file_path):
            click.echo(f"Success: Pulled path=[{file_path}] from digest=" f"[{digest.hash}/{digest.size_bytes}]")
        else:
            click.echo(f'Error: Failed pulling "{file_path}"', err=True)


@cli.command("download-dir", short_help="Download a directory from the CAS server.")
@click.argument("digest-string", nargs=1, type=click.STRING, required=True)
@click.argument("directory-path", nargs=1, type=click.Path(exists=False), required=True)
@click.option("--verify", is_flag=True, show_default=True, help="Check downloaded directory's integrity.")
@pass_context
def download_directory(context: Any, digest_string: str, directory_path: str, verify: bool) -> None:
    if os.path.exists(directory_path):
        if not os.path.isdir(directory_path) or os.listdir(directory_path):
            click.echo("Error: Invalid value, " + f"path=[{directory_path}] already exists.", err=True)
            return

    digest = parse_digest(digest_string)
    assert digest
    try:
        with download(context.channel, instance=context.instance_name) as downloader:
            downloader.download_directory(digest, directory_path)
    except Exception as e:
        click.echo(f"Error: Downloading directory: {e}", err=True)
        sys.exit(-1)

    if verify:
        last_directory_node = None
        for node, _, _ in merkle_tree_maker(directory_path):
            if node.DESCRIPTOR is remote_execution_pb2.DirectoryNode.DESCRIPTOR:
                last_directory_node = node
        if not last_directory_node or last_directory_node.digest != digest:
            click.echo(f"Error: Failed to verify path=[{directory_path}]", err=True)
            return

    if os.path.isdir(directory_path):
        click.echo(f"Success: Pulled path=[{directory_path}] from digest=" f"[{digest.hash}/{digest.size_bytes}]")
    else:
        click.echo(f"Error: Failed pulling path=[{directory_path}]", err=True)



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add Job.Action column

Revision ID: 2b49634f4459
Revises: 38340cc0cfb7
Create Date: 2020-10-29 15:09:49.583803

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2b49634f4459"
down_revision = "38340cc0cfb7"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("jobs", sa.Column("action", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("jobs", "action")
    # ### end Alembic commands ###



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add instance name to jobs

Revision ID: d4efbcc698ca
Revises: afa46425330d
Create Date: 2023-12-11 11:57:45.292788

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d4efbcc698ca"
down_revision = "bc324dfd3610"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("jobs", sa.Column("instance_name", sa.String()))
    op.create_index(op.f("ix_jobs_instance_name"), "jobs", ["instance_name"], unique=False)


def downgrade() -> None:
    op.drop_index(op.f("ix_jobs_instance_name"), table_name="jobs")
    op.drop_column("jobs", "instance_name")



import re
from dataclasses import dataclass, field
from logging import DEBUG, ERROR, INFO, WARNING
from typing import Any, Callable, Dict

import pytest

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest, FindMissingBlobsRequest
from buildgrid.server.enums import OperationStage
from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)

# This anchor proves that our logger is preserving the stack information correctly.
LOC = r"tests.test_logging:test_logging.py:\d+"


@dataclass
class LoggerTest:
    name: str
    expected: str
    level: int = DEBUG
    method: Callable = LOGGER.info
    exc_info: bool = False
    tags: Dict[str, Any] = field(default_factory=dict)


@pytest.mark.parametrize(
    "tc",
    [
        #
        # Test respects logging levels for debug
        #
        LoggerTest(name="debug", level=DEBUG, method=LOGGER.debug, expected=rf"^DEBUG\s+{LOC} foo$"),
        LoggerTest(name="debug", level=INFO, method=LOGGER.debug, expected=r"^$"),
        LoggerTest(name="debug", level=WARNING, method=LOGGER.debug, expected=r"^$"),
        LoggerTest(name="debug", level=ERROR, method=LOGGER.debug, expected=r"^$"),
        #
        # Test respects logging levels for info
        #
        LoggerTest(name="info", level=DEBUG, method=LOGGER.info, expected=rf"^INFO\s+{LOC} foo$"),
        LoggerTest(name="info", level=INFO, method=LOGGER.info, expected=rf"^INFO\s+{LOC} foo$"),
        LoggerTest(name="info", level=WARNING, method=LOGGER.info, expected=r"^$"),
        LoggerTest(name="info", level=ERROR, method=LOGGER.info, expected=r"^$"),
        #
        # Test respects logging levels for warning
        #
        LoggerTest(name="warning", level=DEBUG, method=LOGGER.warning, expected=rf"^WARNING\s+{LOC} foo$"),
        LoggerTest(name="warning", level=INFO, method=LOGGER.warning, expected=rf"^WARNING\s+{LOC} foo$"),
        LoggerTest(name="warning", level=WARNING, method=LOGGER.warning, expected=rf"^WARNING\s+{LOC} foo$"),
        LoggerTest(name="warning", level=ERROR, method=LOGGER.warning, expected=r"^$"),
        #
        # Test respects logging levels for error
        #
        LoggerTest(name="error", level=DEBUG, method=LOGGER.error, expected=rf"^ERROR\s+{LOC} foo$"),
        LoggerTest(name="error", level=INFO, method=LOGGER.error, expected=rf"^ERROR\s+{LOC} foo$"),
        LoggerTest(name="error", level=WARNING, method=LOGGER.error, expected=rf"^ERROR\s+{LOC} foo$"),
        LoggerTest(name="error", level=ERROR, method=LOGGER.error, expected=rf"^ERROR\s+{LOC} foo$"),
        #
        # Test respects logging levels for exception
        #
        LoggerTest(
            name="exception", level=DEBUG, method=LOGGER.exception, expected=rf"^ERROR\s+{LOC} foo\nNoneType: None$"
        ),
        LoggerTest(
            name="exception", level=INFO, method=LOGGER.exception, expected=rf"^ERROR\s+{LOC} foo\nNoneType: None$"
        ),
        LoggerTest(
            name="exception", level=WARNING, method=LOGGER.exception, expected=rf"^ERROR\s+{LOC} foo\nNoneType: None$"
        ),
        LoggerTest(
            name="exception", level=ERROR, method=LOGGER.exception, expected=rf"^ERROR\s+{LOC} foo\nNoneType: None$"
        ),
        #
        # Test respects logging levels for exc_info on all levels
        #
        LoggerTest(
            name="exc_info",
            method=LOGGER.debug,
            exc_info=True,
            expected=rf"^DEBUG\s+{LOC} foo\nNoneType: None$",
        ),
        LoggerTest(
            name="exc_info",
            method=LOGGER.info,
            exc_info=True,
            expected=rf"^INFO\s+{LOC} foo\nNoneType: None$",
        ),
        LoggerTest(
            name="exc_info",
            method=LOGGER.warning,
            exc_info=True,
            expected=rf"^WARNING\s+{LOC} foo\nNoneType: None$",
        ),
        LoggerTest(
            name="exc_info",
            method=LOGGER.error,
            exc_info=True,
            expected=rf"^ERROR\s+{LOC} foo\nNoneType: None$",
        ),
        #
        # Test encoding of value types
        #
        LoggerTest(
            name="None",
            tags=dict(foo=None),
            expected=rf"^INFO\s+{LOC} foo foo=\"\"$",
        ),
        LoggerTest(
            name="int",
            tags=dict(foo=1234),
            expected=rf"^INFO\s+{LOC} foo foo=1234$",
        ),
        LoggerTest(
            name="float",
            tags=dict(foo=1234.1234),
            expected=rf"^INFO\s+{LOC} foo foo=1234.12$",
        ),
        LoggerTest(
            name="Digest",
            tags=dict(foo=Digest(hash="deadbeef", size_bytes=123)),
            expected=rf'^INFO\s+{LOC} foo foo="deadbeef/123"$',
        ),
        LoggerTest(
            name="Message",
            tags=dict(foo=FindMissingBlobsRequest(blob_digests=[Digest(hash="deadbeef", size_bytes=123)])),
            expected=rf'^INFO\s+{LOC} foo foo="blob_digests {{ hash: \\"deadbeef\\" size_bytes: 123 }}"$',
        ),
        LoggerTest(
            name="string",
            tags=dict(foo="bar"),
            expected=rf'^INFO\s+{LOC} foo foo="bar"$',
        ),
        LoggerTest(
            name="string-quotes",
            tags=dict(foo='"bar"'),
            expected=rf'^INFO\s+{LOC} foo foo="\\"bar\\""$',
        ),
        LoggerTest(
            name="enum",
            tags=dict(foo=OperationStage.EXECUTING),
            expected=rf"^INFO\s+{LOC} foo foo=EXECUTING$",
        ),
        # Test tags with non-python variable names
        LoggerTest(
            name="tags",
            tags={"foo.bar": "fizz", "fizz.buzz": 1},
            expected=rf'^INFO\s+{LOC} foo fizz.buzz=1 foo.bar="fizz"$',
        ),
    ],
    ids=lambda tc: tc.name,
)
def test_buildgrid_logger(tc, caplog):
    with caplog.at_level(tc.level):
        if tc.exc_info:
            tc.method("foo", tags=tc.tags, exc_info=tc.exc_info)
        else:
            tc.method("foo", tags=tc.tags)
    assert re.match(tc.expected, caplog.text.strip()), caplog.text



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import inspect
import itertools
from functools import wraps
from typing import Any, Callable, Iterator, TypeVar, cast

import grpc

from buildgrid.server.exceptions import (
    BotSessionCancelledError,
    BotSessionClosedError,
    BotSessionMismatchError,
    DuplicateBotSessionError,
    FailedPreconditionError,
    IncompleteReadError,
    InvalidArgumentError,
    NotFoundError,
    PermissionDeniedError,
    ResourceExhaustedError,
    RetriableError,
    UnknownBotSessionError,
)
from buildgrid.server.logging import buildgrid_logger

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]


LOGGER = buildgrid_logger(__name__)


@contextlib.contextmanager
def _error_context(context: grpc.ServicerContext, unhandled_message: str) -> Iterator[None]:
    try:
        yield

    except BotSessionCancelledError as e:
        LOGGER.info(e)
        context.abort(grpc.StatusCode.CANCELLED, str(e))

    except BotSessionClosedError as e:
        LOGGER.debug(e)
        context.abort(grpc.StatusCode.DATA_LOSS, str(e))

    except IncompleteReadError as e:
        LOGGER.exception(e)
        context.abort(grpc.StatusCode.DATA_LOSS, str(e))

    except ConnectionError as e:
        LOGGER.exception(e)
        context.abort(grpc.StatusCode.UNAVAILABLE, str(e))

    except DuplicateBotSessionError as e:
        LOGGER.info(e)
        context.abort(grpc.StatusCode.ABORTED, str(e))

    except FailedPreconditionError as e:
        LOGGER.error(e)
        context.abort(grpc.StatusCode.FAILED_PRECONDITION, str(e))

    except (InvalidArgumentError, BotSessionMismatchError, UnknownBotSessionError) as e:
        LOGGER.info(e)
        context.abort(grpc.StatusCode.INVALID_ARGUMENT, str(e))

    except NotFoundError as e:
        LOGGER.debug(e)
        context.abort(grpc.StatusCode.NOT_FOUND, str(e))

    except NotImplementedError as e:
        LOGGER.info(e)
        context.abort(grpc.StatusCode.UNIMPLEMENTED, str(e))

    except PermissionDeniedError as e:
        LOGGER.exception(e)
        context.abort(grpc.StatusCode.PERMISSION_DENIED, str(e))

    except RetriableError as e:
        LOGGER.info("Retriable error.", tags=dict(client_retry_delay=e.retry_info.retry_delay))
        context.abort_with_status(e.error_status)

    except ResourceExhaustedError as e:
        LOGGER.exception(e)
        context.abort(grpc.StatusCode.RESOURCE_EXHAUSTED, str(e))

    except Exception as e:
        if context.code() is None:  # type: ignore[attr-defined]
            LOGGER.exception(unhandled_message)
            context.abort(grpc.StatusCode.INTERNAL, str(e))
        raise


def handle_errors(get_printable_request: Callable[[Any], Any] = lambda r: str(r)) -> Callable[[Func], Func]:
    def decorator(f: Func) -> Func:
        @wraps(f)
        def return_wrapper(self: Any, request: Any, context: grpc.ServicerContext) -> Any:
            if isinstance(request, Iterator):
                # Pop the message out to get the instance from it, then and recreate the iterator.
                initial_request = next(request)
                printed_request = get_printable_request(initial_request)
                request = itertools.chain([initial_request], request)
            else:
                printed_request = get_printable_request(request)

            with _error_context(context, f"Unexpected error in {f.__name__}; request=[{printed_request}]"):
                return f(self, request, context)

        @wraps(f)
        def yield_wrapper(self: Any, request: Any, context: grpc.ServicerContext) -> Any:
            if isinstance(request, Iterator):
                # Pop the message out to get the instance from it, then and recreate the iterator.
                initial_request = next(request)
                printed_request = get_printable_request(initial_request)
                request = itertools.chain([initial_request], request)
            else:
                printed_request = get_printable_request(request)

            with _error_context(context, f"Unexpected error in {f.__name__}; request=[{printed_request}]"):
                yield from f(self, request, context)

        if inspect.isgeneratorfunction(f):
            return cast(Func, yield_wrapper)
        return cast(Func, return_wrapper)

    return decorator



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/devtools/build/v1/build_events.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from buildgrid._protos.google.devtools.build.v1 import build_status_pb2 as google_dot_devtools_dot_build_dot_v1_dot_build__status__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+google/devtools/build/v1/build_events.proto\x12\x18google.devtools.build.v1\x1a+google/devtools/build/v1/build_status.proto\x1a\x19google/protobuf/any.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xd0\x0b\n\nBuildEvent\x12.\n\nevent_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x63\n\x1ainvocation_attempt_started\x18\x33 \x01(\x0b\x32=.google.devtools.build.v1.BuildEvent.InvocationAttemptStartedH\x00\x12\x65\n\x1binvocation_attempt_finished\x18\x34 \x01(\x0b\x32>.google.devtools.build.v1.BuildEvent.InvocationAttemptFinishedH\x00\x12L\n\x0e\x62uild_enqueued\x18\x35 \x01(\x0b\x32\x32.google.devtools.build.v1.BuildEvent.BuildEnqueuedH\x00\x12L\n\x0e\x62uild_finished\x18\x37 \x01(\x0b\x32\x32.google.devtools.build.v1.BuildEvent.BuildFinishedH\x00\x12L\n\x0e\x63onsole_output\x18\x38 \x01(\x0b\x32\x32.google.devtools.build.v1.BuildEvent.ConsoleOutputH\x00\x12\x66\n\x19\x63omponent_stream_finished\x18; \x01(\x0b\x32\x41.google.devtools.build.v1.BuildEvent.BuildComponentStreamFinishedH\x00\x12+\n\x0b\x62\x61zel_event\x18< \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x12\x35\n\x15\x62uild_execution_event\x18= \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x12\x32\n\x12source_fetch_event\x18> \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x1aY\n\x18InvocationAttemptStarted\x12\x16\n\x0e\x61ttempt_number\x18\x01 \x01(\x03\x12%\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x1a\x84\x01\n\x19InvocationAttemptFinished\x12@\n\x11invocation_status\x18\x03 \x01(\x0b\x32%.google.devtools.build.v1.BuildStatus\x12%\n\x07\x64\x65tails\x18\x04 \x01(\x0b\x32\x14.google.protobuf.Any\x1a\x36\n\rBuildEnqueued\x12%\n\x07\x64\x65tails\x18\x01 \x01(\x0b\x32\x14.google.protobuf.Any\x1am\n\rBuildFinished\x12\x35\n\x06status\x18\x01 \x01(\x0b\x32%.google.devtools.build.v1.BuildStatus\x12%\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x1a\x86\x01\n\rConsoleOutput\x12;\n\x04type\x18\x01 \x01(\x0e\x32-.google.devtools.build.v1.ConsoleOutputStream\x12\x15\n\x0btext_output\x18\x02 \x01(\tH\x00\x12\x17\n\rbinary_output\x18\x03 \x01(\x0cH\x00\x42\x08\n\x06output\x1a\xc0\x01\n\x1c\x42uildComponentStreamFinished\x12Z\n\x04type\x18\x01 \x01(\x0e\x32L.google.devtools.build.v1.BuildEvent.BuildComponentStreamFinished.FinishType\"D\n\nFinishType\x12\x1b\n\x17\x46INISH_TYPE_UNSPECIFIED\x10\x00\x12\x0c\n\x08\x46INISHED\x10\x01\x12\x0b\n\x07\x45XPIRED\x10\x02\x42\x07\n\x05\x65vent\"\xc8\x01\n\x08StreamId\x12\x10\n\x08\x62uild_id\x18\x01 \x01(\t\x12\x15\n\rinvocation_id\x18\x06 \x01(\t\x12\x44\n\tcomponent\x18\x03 \x01(\x0e\x32\x31.google.devtools.build.v1.StreamId.BuildComponent\"M\n\x0e\x42uildComponent\x12\x15\n\x11UNKNOWN_COMPONENT\x10\x00\x12\x0e\n\nCONTROLLER\x10\x01\x12\n\n\x06WORKER\x10\x02\x12\x08\n\x04TOOL\x10\x03*:\n\x13\x43onsoleOutputStream\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06STDOUT\x10\x01\x12\n\n\x06STDERR\x10\x02\x42\x8b\x01\n\x1c\x63om.google.devtools.build.v1B\x0f\x42uildEventProtoP\x01Z=google.golang.org/genproto/googleapis/devtools/build/v1;build\xf8\x01\x01\xca\x02\x15Google\\Cloud\\Build\\V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.google.devtools.build.v1.build_events_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.devtools.build.v1B\017BuildEventProtoP\001Z=google.golang.org/genproto/googleapis/devtools/build/v1;build\370\001\001\312\002\025Google\\Cloud\\Build\\V1'
  _globals['_CONSOLEOUTPUTSTREAM']._serialized_start=1872
  _globals['_CONSOLEOUTPUTSTREAM']._serialized_end=1930
  _globals['_BUILDEVENT']._serialized_start=179
  _globals['_BUILDEVENT']._serialized_end=1667
  _globals['_BUILDEVENT_INVOCATIONATTEMPTSTARTED']._serialized_start=935
  _globals['_BUILDEVENT_INVOCATIONATTEMPTSTARTED']._serialized_end=1024
  _globals['_BUILDEVENT_INVOCATIONATTEMPTFINISHED']._serialized_start=1027
  _globals['_BUILDEVENT_INVOCATIONATTEMPTFINISHED']._serialized_end=1159
  _globals['_BUILDEVENT_BUILDENQUEUED']._serialized_start=1161
  _globals['_BUILDEVENT_BUILDENQUEUED']._serialized_end=1215
  _globals['_BUILDEVENT_BUILDFINISHED']._serialized_start=1217
  _globals['_BUILDEVENT_BUILDFINISHED']._serialized_end=1326
  _globals['_BUILDEVENT_CONSOLEOUTPUT']._serialized_start=1329
  _globals['_BUILDEVENT_CONSOLEOUTPUT']._serialized_end=1463
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED']._serialized_start=1466
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED']._serialized_end=1658
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED_FINISHTYPE']._serialized_start=1590
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED_FINISHTYPE']._serialized_end=1658
  _globals['_STREAMID']._serialized_start=1670
  _globals['_STREAMID']._serialized_end=1870
  _globals['_STREAMID_BUILDCOMPONENT']._serialized_start=1793
  _globals['_STREAMID_BUILDCOMPONENT']._serialized_end=1870
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from typing import TYPE_CHECKING, Awaitable, Callable, Optional, cast

from aiohttp import web

# TODO Possible upstream fix needed. Module "aiohttp_middlewares" does not explicitly export attribute
# Possibly fixed with dependency update from 1.2.1 to 2.2.0
from aiohttp_middlewares import cors_middleware  # type: ignore[attr-defined]
from aiohttp_middlewares.annotations import Handler, Middleware, UrlCollection

from buildgrid.server.browser import rest_api, utils
from buildgrid.server.browser.utils import ResponseCache

if TYPE_CHECKING:
    from buildgrid.server.app.cli import Context


def serve_client(base_path: str) -> Callable[[web.Request], Awaitable[web.FileResponse]]:
    async def _serve_client(request: web.Request) -> web.FileResponse:
        if not request.match_info["path"]:
            path = os.path.join(base_path, "index.html")
        else:
            path = os.path.join(base_path, request.match_info["path"])
            if not os.path.exists(path):
                path = os.path.join(base_path, "index.html")
        return web.FileResponse(path)

    return _serve_client


def cors_middleware_with_error(origins: UrlCollection, urls: UrlCollection, allow_all: bool = False) -> Middleware:
    assert origins is not None
    assert urls is not None
    middleware = cors_middleware(allow_all=allow_all, origins=origins, urls=urls)

    @web.middleware
    async def _middleware(request: web.Request, handler: Handler) -> web.StreamResponse:
        origin = request.headers.get("Origin")
        headers = utils.get_cors_headers(origin, origins, allow_all)
        try:
            return cast(web.StreamResponse, await middleware(request, handler))

        except web.HTTPNotFound:
            raise web.HTTPNotFound(headers=headers)

        except web.HTTPBadRequest:
            raise web.HTTPBadRequest(headers=headers)

        except web.HTTPInternalServerError:
            raise web.HTTPInternalServerError(headers=headers)

    return _middleware


def create_app(
    context: "Context",
    cache: ResponseCache,
    cors_origins: UrlCollection,
    static_path: Optional[str] = None,
    allow_cancelling_operations: bool = False,
    tarball_dir: Optional[str] = None,
) -> web.Application:
    allow_all = len(cors_origins) == 0
    app = web.Application(
        middlewares=(
            cors_middleware_with_error(allow_all=allow_all, origins=cors_origins, urls=[re.compile(r"^\/api")]),
        )
    )
    routes = [
        web.get("/api/v1/build_events", rest_api.query_build_events_handler(context)),
        web.get("/api/v1/operation_filters", rest_api.get_operation_filters_handler),
        web.get("/api/v1/operations", rest_api.list_operations_handler(context, cache)),
        web.get("/api/v1/operations/{name}", rest_api.get_operation_handler(context, cache)),
        web.get(
            "/api/v1/operations/{name}/request_metadata", rest_api.get_operation_request_metadata_handler(context)
        ),
        web.get("/api/v1/operations/{name}/client_identity", rest_api.get_operation_client_identity_handler(context)),
        web.get("/api/v1/action_results/{hash}/{size_bytes}", rest_api.get_action_result_handler(context, cache)),
        web.get(
            "/api/v1/blobs/{hash}/{size_bytes}",
            rest_api.get_blob_handler(context, cache, allow_all=allow_all, allowed_origins=cors_origins),
        ),
        web.get(
            "/api/v1/tarballs/{hash}_{size_bytes}.tar.gz",
            rest_api.get_tarball_handler(
                context, cache, allow_all=allow_all, allowed_origins=cors_origins, tarball_dir=tarball_dir
            ),
        ),
        web.get("/ws/logstream", rest_api.logstream_handler(context)),
    ]

    if allow_cancelling_operations:
        routes.append(web.delete("/api/v1/operations/{name}", rest_api.cancel_operation_handler(context)))

    if static_path is not None:
        routes.append(web.get("/{path:.*}", serve_client(static_path)))
    app.add_routes(routes)
    return app
